# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import operator

import numpy as np
import pytest
from plum import NotFoundLookupError

from stheno.field import (
    mul,
    add,
    SumElement,
    new,
    get_field,
    broadcast,
    WrappedElement,
    JoinElement,
    Element,
    OneElement,
    ZeroElement,
    ScaledElement,
    ProductElement
)
from stheno.function_field import (
    stretch,
    differentiate,
    shift,
    transform,
    select,
    Function,
    StretchedFunction,
    ShiftedFunction,
    SelectedFunction,
    InputTransformedFunction,
    DerivativeFunction,
    TensorProductFunction,
    OneFunction,
    ZeroFunction,
    _to_list,
    tuple_equal
)
from stheno.function_field import to_tensor
from stheno.kernel import (
    EQ,
    RQ,
    Linear,
    OneKernel,
    ZeroKernel,
    Delta,
    TensorProductKernel,
    Kernel
)
from stheno.matrix import Dense
from stheno.mean import TensorProductMean, ZeroMean, OneMean, Mean
from .util import allclose


def test_corner_cases():
    with pytest.raises(IndexError):
        EQ().stretch(1)[1]
    with pytest.raises(IndexError):
        (EQ() + RQ(1))[2]
    with pytest.raises(RuntimeError):
        mul(1, 1)
    with pytest.raises(RuntimeError):
        add(1, 1)
    with pytest.raises(RuntimeError):
        stretch(1, 1)
    with pytest.raises(RuntimeError):
        transform(1, 1)
    with pytest.raises(RuntimeError):
        differentiate(1, 1)
    with pytest.raises(RuntimeError):
        select(1, 1)
    with pytest.raises(RuntimeError):
        shift(1, 1)
    assert repr(EQ()) == str(EQ())
    assert EQ().__name__ == 'EQ'
    with pytest.raises(NotImplementedError):
        WrappedElement(1).display(1, lambda x: x)
    with pytest.raises(NotImplementedError):
        JoinElement(1, 2).display(1, 2, lambda x: x)


def test_to_tensor():  # From function field.
    assert isinstance(to_tensor(np.array([1, 2])), np.ndarray)
    assert isinstance(to_tensor([1, 2]), np.ndarray)


def test_tuple_equal():
    assert tuple_equal((1,), (1,))
    assert tuple_equal((1, [1]), (1, [1]))
    assert not tuple_equal((1,), ([1],))
    assert not tuple_equal(([1],), ([1], [1]))
    assert tuple_equal(([1],), ([1],))
    assert not tuple_equal(([1],), ([2],))
    assert not tuple_equal(([1], [1, 2]), ([1, 2], [1, 2]))
    assert not tuple_equal(([1], [1, 2]), ([1], [1, 3]))


def test_stretch_shorthand():
    assert str(EQ() > 2) == str(EQ().stretch(2))


def test_equality_field():
    one, zero = OneElement(), ZeroElement()

    assert Element() != Element()

    # Ones and zeros:
    assert one == one
    assert one != zero
    assert zero == zero

    # Scaled elements:
    assert ScaledElement(one, 1) == ScaledElement(one, 1)
    assert ScaledElement(one, 2) != ScaledElement(one, 1)
    assert ScaledElement(zero, 1) != ScaledElement(one, 1)

    # Product elements:
    assert ProductElement(one, zero) == ProductElement(one, zero)
    assert ProductElement(one, zero) == ProductElement(zero, one)
    assert ProductElement(one, zero) != ProductElement(one, one)

    # Sum elements:
    assert SumElement(one, zero) == SumElement(one, zero)
    assert SumElement(one, zero) == SumElement(zero, one)
    assert SumElement(one, zero) != SumElement(one, one)


def test_equality_function_field():
    one, zero = OneFunction(), ZeroFunction()

    def f1(x):
        return x

    def f2(x):
        return x ** 2

    # Test wrapped elements.
    for Type, val1, val2 in [(StretchedFunction, 1, 2),
                             (ShiftedFunction, 1, 2),
                             (SelectedFunction, (1,), (2,)),
                             (InputTransformedFunction, f1, f2),
                             (DerivativeFunction, 1, 2)]:
        assert Type(one, val1) == Type(one, val1)
        assert Type(one, val1) != Type(one, val2)
        assert Type(one, val1) != Type(zero, val1)

    assert TensorProductFunction(f1) == TensorProductFunction(f1)
    assert TensorProductFunction(f1) != TensorProductFunction(f2)


def test_equality_integration():
    # This can be more thorough.
    assert EQ() + EQ() == 2 * EQ()
    assert EQ() != EQ() + 1
    assert EQ().stretch(1).select(0) == EQ().stretch(1).select(0)
    assert Linear() - 1 == Linear() - 1


def test_registrations():
    for x in [Function, Kernel, Mean, Dense]:
        assert get_field.invoke(x)(1) == x

    class MyField(Element):
        pass

    # Test registration of fields and creation of new types.
    with pytest.raises(RuntimeError):
        get_field(MyField())
    get_field.extend(MyField)(lambda _: MyField)
    assert get_field(MyField())
    with pytest.raises(RuntimeError):
        new(MyField(), SumElement)


def test_broadcast():
    assert broadcast(operator.add, (1, 2, 3), (2, 3, 4)) == (3, 5, 7)
    assert broadcast(operator.add, (1,), (2, 3, 4)) == (3, 4, 5)
    assert broadcast(operator.add, (1, 2, 3), (2,)) == (3, 4, 5)
    with pytest.raises(ValueError):
        broadcast(operator.add, (1, 2), (1, 2, 3))
    assert str(EQ().stretch(2).stretch(1, 3)) == 'EQ() > (2, 6)'
    assert str(EQ().stretch(1, 3).stretch(2)) == 'EQ() > (2, 6)'
    assert str(Linear().shift(2).shift(1, 3)) == 'Linear() shift (3, 5)'
    assert str(Linear().shift(1, 3).shift(2)) == 'Linear() shift (3, 5)'


def test_subtraction_and_negation():
    assert str(-EQ()) == '-1 * EQ()'
    assert str(EQ() - EQ()) == '0'
    assert str(RQ(1) - EQ()) == 'RQ(1) + -1 * EQ()'
    assert str(1 - EQ()) == '1 + -1 * EQ()'
    assert str(EQ() - 1) == 'EQ() + -1 * 1'


def test_power():
    with pytest.raises(ValueError):
        EQ() ** -1
    with pytest.raises(NotFoundLookupError):
        EQ() ** .5
    assert str(EQ() ** 0) == '1'
    assert str(EQ() ** 1) == 'EQ()'
    assert str(EQ() ** 2) == 'EQ() * EQ()'
    assert str(EQ() ** 3) == 'EQ() * EQ() * EQ()'


def test_cancellations_zero():
    # With constants:
    assert str(1 * EQ()) == 'EQ()'
    assert str(EQ() * 1) == 'EQ()'
    assert str(0 * EQ()) == '0'
    assert str(EQ() * 0) == '0'
    assert str(0 + EQ()) == 'EQ()'
    assert str(EQ() + 0) == 'EQ()'
    assert str(0 + OneMean()) == '1'
    assert str(OneMean() + 0) == '1'

    # Adding to zero:
    assert str(0 + ZeroKernel()) == '0'
    assert str(ZeroKernel() + 0) == '0'
    assert str(1 + ZeroKernel()) == '1'
    assert str(ZeroKernel() + 1) == '1'
    assert str(2 + ZeroKernel()) == '2 * 1'
    assert str(ZeroKernel() + 2) == '2 * 1'

    # Sums:
    assert str(EQ() + EQ()) == '2 * EQ()'
    assert str(ZeroKernel() + EQ()) == 'EQ()'
    assert str(EQ() + ZeroKernel()) == 'EQ()'
    assert str(ZeroKernel() + ZeroKernel()) == '0'

    # Products:
    assert str(EQ() * EQ()) == 'EQ() * EQ()'
    assert str(ZeroKernel() * EQ()) == '0'
    assert str(EQ() * ZeroKernel()) == '0'
    assert str(ZeroKernel() * ZeroKernel()) == '0'

    # Scales:
    assert str(5 * ZeroKernel()) == '0'
    assert str(ZeroKernel() * 5) == '0'
    assert str(EQ() * 5) == '5 * EQ()'
    assert str(5 * EQ()) == '5 * EQ()'

    # Stretches:
    assert str(ZeroKernel().stretch(5)) == '0'
    assert str(EQ().stretch(5)) == 'EQ() > 5'

    # Periodicisations:
    assert str(ZeroKernel().periodic(5)) == '0'
    assert str(EQ().periodic(5)) == 'EQ() per 5'

    # Reversals:
    assert str(reversed(ZeroKernel())) == '0'
    assert str(reversed(EQ())) == 'EQ()'
    assert str(reversed(Linear())) == 'Reversed(Linear())'

    # Integration:
    assert str(EQ() * EQ() + ZeroKernel() * EQ()) == 'EQ() * EQ()'
    assert str(EQ() * ZeroKernel() + ZeroKernel() * EQ()) == '0'


def test_cancellations_one():
    # Products:
    assert str(EQ() * EQ()) == 'EQ() * EQ()'
    assert str(OneKernel() * EQ()) == 'EQ()'
    assert str(EQ() * OneKernel()) == 'EQ()'
    assert str(OneKernel() * OneKernel()) == '1'


def test_grouping():
    # Scales:
    assert str(5 * EQ()) == '5 * EQ()'
    assert str(5 * (5 * EQ())) == '25 * EQ()'

    # Stretches:
    assert str(EQ().stretch(5)) == 'EQ() > 5'
    assert str(EQ().stretch(5).stretch(5)) == 'EQ() > 25'

    # Shifts:
    assert str(Linear().shift(5)) == 'Linear() shift 5'
    assert str(Linear().shift(5).shift(5)) == 'Linear() shift 10'

    # Products:
    assert str((5 * EQ()) * (5 * EQ())) == '25 * EQ()'
    assert str((5 * (EQ() * EQ())) * (5 * EQ() * EQ())) == '25 * EQ() * EQ()'
    assert str((5 * RQ(1)) * (5 * RQ(2))) == '25 * RQ(1) * RQ(2)'

    # Sums:
    assert str((5 * EQ()) + (5 * EQ())) == '10 * EQ()'
    assert str(EQ() + (5 * EQ())) == '6 * EQ()'
    assert str((5 * EQ()) + EQ()) == '6 * EQ()'
    assert str((EQ() + EQ())) == '2 * EQ()'
    assert str((5 * (EQ() * EQ())) + (5 * (EQ() * EQ()))) == '10 * EQ() * EQ()'
    assert str((5 * RQ(1)) + (5 * RQ(2))) == '5 * RQ(1) + 5 * RQ(2)'

    # Reversal:
    assert str(reversed(Linear() + EQ())) == 'Reversed(Linear()) + EQ()'
    assert str(reversed(Linear() * EQ())) == 'Reversed(Linear()) * EQ()'


def test_parentheses():
    assert str((reversed(Linear() * Linear() +
                         2 * EQ().stretch(1).periodic(2) +
                         RQ(3).periodic(4))) * (EQ().stretch(1) + EQ())) == \
           '(Reversed(Linear()) * Reversed(Linear()) + ' \
           '2 * ((EQ() > 1) per 2) + ' \
           'RQ(3) per 4) * (EQ() > 1 + EQ())'


def test_terms():
    k = EQ() + EQ() * Linear() + RQ(1) * RQ(2) + Delta()
    assert k.num_terms == 4
    assert str(k.term(0)) == 'EQ()'
    assert str(k.term(1)) == 'EQ() * Linear()'
    assert str(k.term(2)) == 'RQ(1) * RQ(2)'
    assert str(k.term(3)) == 'Delta()'
    with pytest.raises(IndexError):
        k.term(4)
    with pytest.raises(IndexError):
        EQ().term(1)


def test_factors():
    k = EQ() * Linear()
    assert k.num_factors == 2
    assert str(k.factor(0)) == 'EQ()'
    assert str(k.factor(1)) == 'Linear()'
    with pytest.raises(IndexError):
        k.factor(2)

    k = (EQ() + EQ()) * Delta() * (RQ(1) + Linear())
    assert k.num_factors == 4
    assert str(k.factor(0)) == '2'
    assert str(k.factor(1)) == 'EQ()'
    assert str(k.factor(2)) == 'Delta()'
    assert str(k.factor(3)) == 'RQ(1) + Linear()'
    with pytest.raises(IndexError):
        k.factor(4)
    with pytest.raises(IndexError):
        EQ().factor(1)


def test_indexing():
    assert str((5 * EQ())[0]) == 'EQ()'
    assert str((EQ() + RQ(1.0))[0]) == 'EQ()'
    assert str((EQ() + RQ(1.0))[1]) == 'RQ(1.0)'
    assert str((EQ() * RQ(1.0))[0]) == 'EQ()'
    assert str((EQ() * RQ(1.0))[1]) == 'RQ(1.0)'


def test_shifting():
    # Kernels:
    assert str(ZeroKernel().shift(5)) == '0'
    assert str(EQ().shift(5)) == 'EQ()'
    assert str(Linear().shift(5)) == 'Linear() shift 5'
    assert str((5 * EQ()).shift(5)) == '5 * EQ()'
    assert str((5 * Linear()).shift(5)) == '(5 * Linear()) shift 5'

    # Means:
    def mean(x): return x

    m = TensorProductMean(mean)
    assert str(ZeroMean().shift(5)) == '0'
    assert str(m.shift(5)) == 'mean shift 5'
    assert str(m.shift(5).shift(5)) == 'mean shift 10'
    assert str((5 * m).shift(5)) == '(5 * mean) shift 5'


def test_selection():
    allclose(_to_list((1, 2)), [1, 2])
    allclose(_to_list([1, 2]), [1, 2])
    allclose(_to_list(np.array([1, 2])), [1, 2])
    allclose(_to_list(1), [1])
    with pytest.raises(ValueError):
        _to_list(np.ones((1, 1)))

    assert str(EQ().select(0)) == 'EQ() : [0]'
    assert str(EQ().select([0, 2])) == 'EQ() : [0, 2]'
    assert str(EQ().select(0, 2)) == 'EQ() : ([0], [2])'
    assert str(EQ().select([0], 2)) == 'EQ() : ([0], [2])'
    assert str(EQ().select(0, [2])) == 'EQ() : ([0], [2])'
    assert str(EQ().select([0, 1], [2])) == 'EQ() : ([0, 1], [2])'
    assert str(EQ().select(None, [2])) == 'EQ() : (None, [2])'
    assert str(EQ().select(None)) == 'EQ() : None'
    assert str(EQ().select(None, None)) == 'EQ() : (None, None)'
    assert str(EQ().select([1], None)) == 'EQ() : ([1], None)'

    assert str(ZeroKernel().select(0)) == '0'
    assert str(OneMean().select(0)) == '1'


def test_input_transform():
    assert str(EQ().transform(lambda x: x)) == 'EQ() transform <lambda>'
    assert str(EQ().transform(lambda x: x, lambda x: x)) == \
           'EQ() transform (<lambda>, <lambda>)'

    assert str(ZeroKernel().transform(lambda x: x)) == '0'
    assert str(OneMean().transform(lambda x: x)) == '1'


def test_derivative():
    assert str(EQ().diff(0)) == 'd(0) EQ()'
    assert str(EQ().diff(0, 1)) == 'd(0, 1) EQ()'

    assert str(ZeroKernel().diff(0)) == '0'
    assert str(OneKernel().diff(0)) == '0'

    assert str(ZeroMean().diff(0)) == '0'
    assert str(OneMean().diff(0)) == '0'


def test_function():
    def f():
        pass

    assert str(ZeroKernel() * f) == '0'
    assert str(f * ZeroKernel()) == '0'
    assert str(OneKernel() * f) == 'f'
    assert str(f * OneKernel()) == 'f'
    assert str(EQ() * f) == 'EQ() * f'
    assert str(f * EQ()) == 'f * EQ()'
    assert str((EQ() * EQ()) * f) == 'EQ() * EQ() * f'
    assert str(f * (EQ() * EQ())) == 'f * EQ() * EQ()'
    assert str((5 * EQ()) * f) == '5 * EQ() * f'
    assert str(f * (5 * EQ())) == '5 * f * EQ()'

    assert str(ZeroKernel() + f) == 'f'
    assert str(f + ZeroKernel()) == 'f'
    assert str(OneKernel() + f) == '1 + f'
    assert str(f + OneKernel()) == 'f + 1'
    assert str(EQ() + f) == 'EQ() + f'
    assert str(f + EQ()) == 'f + EQ()'
    assert str((EQ() + RQ(1)) + f) == 'EQ() + RQ(1) + f'
    assert str(f + (EQ() + RQ(1))) == 'f + EQ() + RQ(1)'
    assert str((5 + EQ()) + f) == '5 * 1 + EQ() + f'
    assert str(f + (5 + EQ())) == 'f + 5 * 1 + EQ()'

    assert str(OneKernel() * f) == 'f'
    assert str(OneKernel() * TensorProductKernel((lambda x, c: x), f)) == \
           '(<lambda> x f)'
    assert str(OneKernel() * TensorProductKernel(f, (
        lambda x, c: x))) == '(f x <lambda>)'


def test_formatting():
    assert (2 * EQ().stretch(3)).display(lambda x: x ** 2) == '4 * (EQ() > 9)'



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import numpy as np
import pytest
from lab import B
from scipy.stats import multivariate_normal

from stheno.graph import GP
from stheno.matrix import UniformlyDiagonal, Diagonal, dense
from stheno.random import Normal, Normal1D, RandomVector
from .util import allclose


def test_normal():
    mean = np.random.randn(3, 1)
    chol = np.random.randn(3, 3)
    var = chol.dot(chol.T)

    dist = Normal(var, mean)
    dist_sp = multivariate_normal(mean[:, 0], var)

    # Test second moment.
    allclose(dist.m2, var + mean.dot(mean.T))

    # Test marginals.
    marg_mean, lower, upper = dist.marginals()
    allclose(mean.squeeze(), marg_mean)
    allclose(lower, marg_mean - 2 * np.diag(var) ** .5)
    allclose(upper, marg_mean + 2 * np.diag(var) ** .5)

    # Test `logpdf` and `entropy`.
    for _ in range(5):
        x = np.random.randn(3, 10)
        allclose(dist.logpdf(x), dist_sp.logpdf(x.T), desc='logpdf')
        allclose(dist.entropy(), dist_sp.entropy(), desc='entropy')

    # Test the the output of `logpdf` is flattened appropriately.
    assert np.shape(dist.logpdf(np.ones((3, 1)))) == ()
    assert np.shape(dist.logpdf(np.ones((3, 2)))) == (2,)

    # Test KL with Monte Carlo estimate.
    mean2 = np.random.randn(3, 1)
    chol2 = np.random.randn(3, 3)
    var2 = chol2.dot(chol2.T)
    dist2 = Normal(var2, mean2)
    samples = dist.sample(50000)
    kl_est = np.mean(dist.logpdf(samples)) - np.mean(dist2.logpdf(samples))
    kl = dist.kl(dist2)
    assert np.abs(kl_est - kl) / np.abs(kl) < 5e-2, 'kl sampled'


def test_normal_comparison():
    # Compare a diagonal normal and dense normal.
    mean = np.random.randn(3, 1)
    var_diag = np.random.randn(3) ** 2
    var = np.diag(var_diag)
    dist1 = Normal(var, mean)
    dist2 = Normal(Diagonal(var_diag), mean)
    samples = dist1.sample(100)
    allclose(dist1.logpdf(samples), dist2.logpdf(samples), desc='logpdf')
    allclose(dist1.entropy(), dist2.entropy(), desc='entropy')
    allclose(dist1.kl(dist2), 0.)
    allclose(dist1.kl(dist1), 0.)
    allclose(dist2.kl(dist2), 0.)
    allclose(dist2.kl(dist1), 0.)
    assert dist1.w2(dist1) <= 1e-3
    assert dist1.w2(dist2) <= 1e-3
    assert dist2.w2(dist1) <= 1e-3
    assert dist2.w2(dist2) <= 1e-3

    # Check a uniformly diagonal normal and dense normal.
    mean = np.random.randn(3, 1)
    var_diag_scale = np.random.randn() ** 2
    var = np.eye(3) * var_diag_scale
    dist1 = Normal(var, mean)
    dist2 = Normal(UniformlyDiagonal(var_diag_scale, 3), mean)
    samples = dist1.sample(100)
    allclose(dist1.logpdf(samples), dist2.logpdf(samples), desc='logpdf')
    allclose(dist1.entropy(), dist2.entropy(), desc='entropy')
    allclose(dist1.kl(dist2), 0.)
    allclose(dist1.kl(dist1), 0.)
    allclose(dist2.kl(dist2), 0.)
    allclose(dist2.kl(dist1), 0.)
    assert dist1.w2(dist1) <= 1e-3
    assert dist1.w2(dist2) <= 1e-3
    assert dist2.w2(dist1) <= 1e-3
    assert dist2.w2(dist2) <= 1e-3


def test_normal_sampling():
    # Test sampling and dtype conversion.
    dist = Normal(3 * np.eye(200, dtype=np.integer))
    assert np.abs(np.std(dist.sample(1000)) ** 2 - 3) <= 5e-2, 'full'
    assert np.abs(np.std(dist.sample(1000, noise=2)) ** 2 - 5) <= 5e-2, 'full 2'

    dist = Normal(Diagonal(3 * np.ones(200, dtype=np.integer)))
    assert np.abs(np.std(dist.sample(1000)) ** 2 - 3) <= 5e-2, 'diag'
    assert np.abs(np.std(dist.sample(1000, noise=2)) ** 2 - 5) <= 5e-2, 'diag 2'

    dist = Normal(UniformlyDiagonal(3, 200))
    assert np.abs(np.std(dist.sample(1000)) ** 2 - 3) <= 5e-2, 'unif'
    assert np.abs(np.std(dist.sample(1000, noise=2)) ** 2 - 5) <= 5e-2, 'unif 2'

    # Test `__str__` and `__repr__`.
    assert str(dist) == RandomVector.__str__(dist)
    assert repr(dist) == RandomVector.__repr__(dist)

    # Test zero mean determination.
    assert Normal(np.eye(3))._zero_mean
    assert not Normal(np.eye(3), np.random.randn(3, 1))._zero_mean

    x = np.random.randn(3)
    assert GP(1)(x)._zero_mean
    assert not GP(1, 1)(x)._zero_mean
    assert GP(1, 0)(x)._zero_mean


def test_normal_1d():
    # Test broadcasting.
    d = Normal1D(1, 0)
    assert type(d.var) == UniformlyDiagonal
    assert B.shape(d.var) == (1, 1)
    assert B.shape(d.mean) == (1, 1)

    d = Normal1D(1, np.array([0, 0, 0]))
    assert type(d.var) == UniformlyDiagonal
    assert B.shape(d.var) == (3, 3)
    assert B.shape(d.mean) == (3, 1)

    d = Normal1D(np.array([1, 2, 3]), 0)
    assert type(d.var) == Diagonal
    assert B.shape(d.var) == (3, 3)
    assert B.shape(d.mean) == (3, 1)

    d = Normal1D(np.array([1, 2, 3]), np.array([0, 0, 0]))
    assert type(d.var) == Diagonal
    assert B.shape(d.var) == (3, 3)
    assert B.shape(d.mean) == (3, 1)

    d = Normal1D(1)
    assert type(d.var) == UniformlyDiagonal
    assert B.shape(d.var) == (1, 1)
    assert B.shape(d.mean) == (1, 1)

    d = Normal1D(np.array([1, 2, 3]))
    assert type(d.var) == Diagonal
    assert B.shape(d.var) == (3, 3)
    assert B.shape(d.mean) == (3, 1)

    with pytest.raises(ValueError):
        Normal1D(np.eye(3))
    with pytest.raises(ValueError):
        Normal1D(np.eye(3), 0)
    with pytest.raises(ValueError):
        Normal1D(1, np.ones((3, 1)))
    with pytest.raises(ValueError):
        Normal1D(np.array([1, 2]),
                 np.ones((3, 1)))


def test_normal_arithmetic():
    chol = np.random.randn(3, 3)
    dist = Normal(chol.dot(chol.T), np.random.randn(3, 1))
    chol = np.random.randn(3, 3)
    dist2 = Normal(chol.dot(chol.T), np.random.randn(3, 1))

    A = np.random.randn(3, 3)
    a = np.random.randn(1, 3)
    b = 5.

    # Test matrix multiplication.
    allclose((dist.rmatmul(a)).mean, dist.mean.dot(a))
    allclose((dist.rmatmul(a)).var, a.dot(dense(dist.var)).dot(a.T))
    allclose((dist.lmatmul(A)).mean, A.dot(dist.mean))
    allclose((dist.lmatmul(A)).var, A.dot(dense(dist.var)).dot(A.T))

    # Test multiplication.
    allclose((dist * b).mean, dist.mean * b)
    allclose((dist * b).var, dist.var * b ** 2)
    allclose((b * dist).mean, dist.mean * b)
    allclose((b * dist).var, dist.var * b ** 2)
    with pytest.raises(NotImplementedError):
        dist.__mul__(dist)
    with pytest.raises(NotImplementedError):
        dist.__rmul__(dist)

    # Test addition.
    allclose((dist + dist2).mean, dist.mean + dist2.mean)
    allclose((dist + dist2).var, dist.var + dist2.var)
    allclose((dist.__add__(b)).mean, dist.mean + b)
    allclose((dist.__radd__(b)).mean, dist.mean + b)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from lab import B

from .matrix import *
from .input import *
from .field import *
from .util import *
from .mean import *
from .momean import *
from .kernel import *
from .mokernel import *
from .random import *
from .eis import *
from .lazy import *
from .graph import *
from .autograd import *


# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import logging
from numbers import Number
from string import ascii_lowercase

import numpy as np
from lab import B
from plum import Referentiable, Self, Dispatcher, add_conversion_method, \
    add_promotion_rule

from .field import Element, OneElement, ZeroElement, mul, add, get_field

__all__ = ['matrix',
           'Dense',
           'LowRank',
           'Diagonal',
           'UniformlyDiagonal',
           'One',
           'Zero',
           'dense',
           'Woodbury',
           'Constant']

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


class Dense(Element, Referentiable):
    """Symmetric positive-definite matrix.

    Args:
        mat (tensor): Symmetric positive-definite matrix.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, mat):
        self.mat = mat

        # Caching:
        self.cholesky = None
        self.logdet = None
        self.root = None
        self.inverse = None

    def __neg__(self):
        return mul(B.cast(B.dtype(self), -1), self)

    def __div__(self, other):
        return B.divide(self, other)

    def __truediv__(self, other):
        return B.divide(self, other)

    def __rdiv__(self, other):
        return B.divide(other, self)

    def __rtruediv__(self, other):
        return B.divide(other, self)

    def __getitem__(self, item):
        return dense(self)[item]

    @property
    def T(self):
        return B.transpose(self)

    def __matmul__(self, other):
        return B.matmul(self, other)

    @_dispatch(Self)
    def __eq__(self, other):
        return dense(self) == dense(other)


# Register field.
@get_field.extend(Dense)
def get_field(a): return Dense


class Diagonal(Dense, Referentiable):
    """Diagonal symmetric positive-definite matrix.

    Args:
        diag (vector): Diagonal of the matrix.
        rows (scalar, optional): Number of rows. Defaults to the length of the
            diagonal.
        cols (scalar, optional): Number of columns. Defaults to the number of
            rows.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, diag, rows=None, cols=None):
        Dense.__init__(self, None)
        self.diag = diag
        self.rows = B.shape(diag)[0] if rows is None else rows
        self.cols = self.rows if cols is None else cols

    @_dispatch(Self)
    def __eq__(self, other):
        return B.diag(self) == B.diag(other)


class UniformlyDiagonal(Diagonal, Referentiable):
    """Uniformly diagonal symmetric positive-definite matrix.

    Args:
        diag_scale (scalar): Scale of the diagonal of the matrix.
        n (int): Length of the diagonal.
        rows (scalar, optional): Number of rows. Defaults to the length of the
            diagonal.
        cols (scalar, optional): Number of columns. Defaults to the number of
            rows.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, diag_scale, n, rows=None, cols=None):
        diag = diag_scale * B.ones(B.dtype(diag_scale), n)
        Diagonal.__init__(self, diag=diag, rows=rows, cols=cols)

    @classmethod
    def from_(cls, diag_scale, ref):
        return cls(B.cast(B.dtype(ref), diag_scale),
                   B.diag_len(ref),
                   *B.shape(ref))


class LowRank(Dense, Referentiable):
    """Low-rank symmetric positive-definite matrix.

    The low-rank matrix is constructed via `left diag(scales) transpose(right)`.

    Args:
        left (tensor): Left part of the matrix.
        right (tensor, optional): Right part of the matrix. Defaults to `left`.
        scales (tensor, optional): Scaling of the outer products. Defaults to
            ones.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, left, right=None, middle=None):
        Dense.__init__(self, None)
        self.left = left
        self.right = left if right is None else right
        if middle is None:
            self.middle = B.eye(B.dtype(self.left), B.shape(self.left)[1])
        else:
            self.middle = middle

        # Shorthands:
        self.l = self.left
        self.r = self.right
        self.m = self.middle

    @_dispatch(Self)
    def __eq__(self, other):
        return (self.left == other.left,
                self.middle == other.middle,
                self.right == other.right)


class Constant(LowRank, Referentiable):
    """Constant symmetric positive-definite matrix.

    Args:
        constant (scalar): Constant of the matrix.
        rows (scalar): Number of rows.
        cols (scalar, optional): Number of columns. Defaults to the number of
            rows.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, constant, rows, cols=None):
        self.constant = constant
        self.rows = rows
        self.cols = rows if cols is None else cols

        # Construct and initialise the low-rank representation.
        left = B.ones(B.dtype(self.constant), self.rows, 1)
        if self.rows is self.cols:
            right = left
        else:
            right = B.ones(B.dtype(self.constant), self.cols, 1)
        middle = B.expand_dims(B.expand_dims(self.constant, axis=0), axis=0)
        LowRank.__init__(self, left=left, right=right, middle=middle)

    @classmethod
    def from_(cls, constant, ref):
        return cls(B.cast(B.dtype(ref), constant), *B.shape(ref))

    def __eq__(self, other):
        return B.shape(self) == B.shape(other) \
               and self.constant == other.constant


class One(Constant, OneElement, Referentiable):
    """Dense matrix full of ones.

    Args:
        dtype (dtype): Data type.
        rows (scalar): Number of rows.
        cols (scalar, optional): Number of columns. Defaults to the number of
            rows.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, dtype, rows, cols=None):
        Constant.__init__(self, B.cast(dtype, 1), rows=rows, cols=cols)

    @classmethod
    def from_(cls, ref):
        return cls(B.dtype(ref), *B.shape(ref))

    @_dispatch(Self)
    def __eq__(self, other):
        return B.shape(self) == B.shape(other)


class Zero(Constant, ZeroElement, Referentiable):
    """Dense matrix full of zeros.

    Args:
        dtype (dtype): Data type.
        rows (scalar): Number of rows.
        cols (scalar, optional): Number of columns. Defaults to the number of
            rows.
    """
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.DType, [object])
    def __init__(self, dtype, rows, cols=None):
        Constant.__init__(self, B.cast(dtype, 0), rows=rows, cols=cols)

    @classmethod
    def from_(cls, ref):
        return cls(B.dtype(ref), *B.shape(ref))

    @_dispatch(Self)
    def __eq__(self, other):
        return B.shape(self) == B.shape(other)


class Woodbury(Dense, Referentiable):
    """Sum of a low-rank and diagonal symmetric positive-definite matrix.

    Args:
        diag (:class:`.matrix.Diagonal`): Diagonal part.
        lr (:class:`.matrix.LowRank`): Low-rank part.
        lr_pd (bool, optional): Specify that the low-rank part is PD. Defaults
            to `True`.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, diag, lr, lr_pd=True):
        Dense.__init__(self, None)
        self.diag = diag
        self.lr = lr
        self.lr_pd = lr_pd

        # Caching:
        self.schur = None

    @_dispatch(Self)
    def __eq__(self, other):
        return self.lr == other.lr, self.diag == other.diag


# Conveniently make identity matrices.

@B.eye.extend(Dense)
def eye(a): return B.eye(B.dtype(a), *B.shape(a))


# Conversion between matrices and dense matrices:

@_dispatch(Dense)
def matrix(a):
    """Matrix as `Dense`.

    Args:
        a (tensor): Matrix to type.

    Returns:
        :class:`.matrix.Dense`: Matrix as `Dense`.
    """
    return a


@_dispatch(B.Numeric)
def matrix(a): return Dense(a)


@_dispatch(Dense)
def dense(a):
    """`Dense` as matrix.

    Args:
        a (:class:`.matrix.Dense`): `Dense` to unwrap.

    Returns:
        tensor: `Dense` as matrix.
    """
    return a.mat


@_dispatch(B.Numeric)
def dense(a): return a


@_dispatch(Diagonal)
def dense(a): return B.diag(B.diag(a), *B.shape(a))


@_dispatch(LowRank)
def dense(a): return B.matmul(a.left, a.middle, a.right, tr_c=True)


@_dispatch(Woodbury)
def dense(a): return dense(a.lr) + dense(a.diag)


# Construct a matrix from blocks.

def block_matrix(*rows):
    """Construct a matrix from its blocks, preserving structure when possible.

    Args:
        *rows (list): Rows of the block matrix.

    Returns:
        :class:`.matrix.Dense`: Assembled matrix as structured as possible.
    """
    # If the blocks form a grid, then might know how to preserve structure.
    grid = True

    # Check that all shapes line up.
    for r in range(1, len(rows)):
        for c in range(1, len(rows[0])):
            block_shape = B.shape(rows[r][c])

            # The previous block in the row must have an equal
            # number of rows.
            if block_shape[0] != B.shape(rows[r][c - 1])[0]:
                grid = False

            # The previous block in the column must have an equal
            # number of columns.
            if block_shape[1] != B.shape(rows[r - 1][c])[1]:
                grid = False

    if grid:
        # We have a grid! First, determine the resulting shape.
        grid_rows = _builtin_sum([B.shape(row[0])[0] for row in rows])
        grid_cols = _builtin_sum([B.shape(K)[1] for K in rows[0]])

        # Check whether the result is just zeros.
        if all([all([isinstance(K, Zero) for K in row]) for row in rows]):
            return Zero(B.dtype(rows[0][0]),
                        rows=grid_rows,
                        cols=grid_cols)

        # Check whether the result is just ones.
        if all([all([isinstance(K, One) for K in row]) for row in rows]):
            return One(B.dtype(rows[0][0]),
                       rows=grid_rows,
                       cols=grid_cols)

        # Check whether the result is diagonal.
        diagonal = True
        diagonal_blocks = []

        for r in range(len(rows)):
            for c in range(len(rows[0])):
                block_shape = B.shape(rows[r][c])
                if r == c:
                    # Keep track of all the diagonal blocks.
                    diagonal_blocks.append(rows[r][c])

                    # All blocks on the diagonal must be diagonal or zero.
                    if not isinstance(rows[r][c], (Diagonal, Zero)):
                        diagonal = False

                    # All blocks on the diagonal must be square.
                    if not block_shape[0] == block_shape[1]:
                        diagonal = False

                # All blocks not on the diagonal must be zero.
                if r != c and not isinstance(rows[r][c], Zero):
                    diagonal = False

        if diagonal:
            return Diagonal(B.concat(*[B.diag(K) for K in diagonal_blocks],
                                     axis=0),
                            rows=grid_rows,
                            cols=grid_cols)

    # We could not preserve any structure. Simply concatenate them all
    # densely.
    return Dense(B.concat2d(*[[dense(K) for K in row] for row in rows]))


B.block_matrix = block_matrix  # Record in LAB.


# Get data type of matrices.

@B.dtype.extend(Dense)
def dtype(a): return B.dtype(dense(a))


@B.dtype.extend(Diagonal)
def dtype(a): return B.dtype(a.diag)


@B.dtype.extend(LowRank)
def dtype(a): return B.dtype(a.left)


@B.dtype.extend(Constant)
def dtype(a): return B.dtype(a.constant)


@B.dtype.extend(Woodbury)
def dtype(a): return B.dtype(a.lr)


# Sum over matrices.

_builtin_sum = sum  # Save built-in function.


@B.sum.extend(Dense, [object])
def sum(a, axis=None): return B.sum(dense(a), axis=axis)


@B.sum.extend(Diagonal, [object])
def sum(a, axis=None):
    # Efficiently handle a number of common cases.
    if axis is None:
        return B.sum(B.diag(a))
    elif axis is 0:
        return B.concat(B.diag(a),
                        B.zeros(B.dtype(a), B.shape(a)[1] - B.diag_len(a)),
                        axis=0)
    elif axis is 1:
        return B.concat(B.diag(a),
                        B.zeros(B.dtype(a), B.shape(a)[0] - B.diag_len(a)),
                        axis=0)
    else:
        # Fall back to generic implementation.
        return B.sum.invoke(Dense)(a, axis=axis)


@B.sum.extend(LowRank, [object])
def sum(a, axis=None):
    # Efficiently handle a number of common cases.
    if axis is None:
        left = B.expand_dims(B.sum(a.left, axis=0), axis=0)
        right = B.expand_dims(B.sum(a.right, axis=0), axis=1)
        return B.matmul(left, a.middle, right)[0, 0]
    elif axis is 0:
        left = B.expand_dims(B.sum(a.left, axis=0), axis=0)
        return B.matmul(left, a.middle, a.right, tr_c=True)[0, :]
    elif axis is 1:
        right = B.expand_dims(B.sum(a.right, axis=0), axis=1)
        return B.matmul(a.left, a.middle, right)[:, 0]
    else:
        # Fall back to generic implementation.
        return B.sum.invoke(Dense)(a, axis=axis)


@B.sum.extend(Woodbury, [object])
def sum(a, axis=None): return B.sum(a.lr, axis=axis) + B.sum(a.diag, axis=axis)


# Take traces of matrices.

@B.trace.extend(Dense)
def trace(a): return B.trace(dense(a))


# Get the length of the diagonal of a matrix.

def diag_len(a):
    """Get the length of the diagonal of a matrix.

    Args:
        a (tensor): Matrix of which to get the length of the diagonal.

    Returns:
        tensor: Length of the diagonal of `a`.
    """
    return B.minimum(*B.shape(a))


B.diag_len = diag_len  # Record in LAB.


# Get diagonals of matrices or create diagonal matrices.

@B.diag.extend(Dense)
def diag(a): return B.diag(dense(a))


@B.diag.extend(Diagonal)
def diag(a):
    # Append zeros or remove elements as necessary.
    diag_len = B.diag_len(a)
    extra_zeros = B.maximum(diag_len - B.shape(a.diag)[0], 0)
    return B.concat(a.diag[:diag_len], B.zeros(B.dtype(a), extra_zeros), axis=0)


@B.diag.extend(LowRank)
def diag(a):
    # The matrix might be non-square, so handle that.
    diag_len = B.diag_len(a)
    return B.sum(B.matmul(a.left, a.middle)[:diag_len, :] *
                 a.right[:diag_len, :], axis=1)


@B.diag.extend(Constant)
def diag(a): return a.constant * B.ones(B.dtype(a), B.diag_len(a))


@B.diag.extend(Woodbury)
def diag(a): return B.diag(a.lr) + B.diag(a.diag)


@B.diag.extend(B.Numeric, B.Numeric, [B.Numeric])
def diag(diag, rows, cols=None):
    cols = rows if cols is None else cols

    # Cut the diagonal to accommodate the size.
    diag = diag[:B.minimum(rows, cols)]
    diag_len, dtype = B.shape(diag)[0], B.dtype(diag)

    # PyTorch incorrectly handles dimensions of size 0. Therefore, if the
    # numbers of extra columns and rows are `Number`s, which will be the case if
    # PyTorch is the backend, then perform a check to prevent appending tensors
    # with dimensions of size 0.

    # Start with just a diagonal matrix.
    res = B.diag(diag)

    # Pad extra columns if necessary.
    extra_cols = cols - diag_len
    if not (isinstance(extra_cols, Number) and extra_cols == 0):
        zeros = B.zeros(dtype, diag_len, extra_cols)
        res = B.concat(B.diag(diag), zeros, axis=1)

    # Pad extra rows if necessary.
    extra_rows = rows - diag_len
    if not (isinstance(extra_rows, Number) and extra_rows == 0):
        zeros = B.zeros(dtype, extra_rows, diag_len + extra_cols)
        res = B.concat(res, zeros, axis=0)

    return res


# Cholesky decompose matrices.

@B.cholesky.extend(Dense)
def cholesky(a):
    if a.cholesky is None:
        a.cholesky = B.cholesky(B.reg(dense(a)))
    return a.cholesky


@B.cholesky.extend(Diagonal)
def cholesky(a):
    # NOTE: Assumes that the matrix is symmetric.
    if a.cholesky is None:
        a.cholesky = Diagonal(a.diag ** .5)
    return a.cholesky


@B.cholesky.extend(LowRank)
def cholesky(a):
    # NOTE: Assumes that the matrix is symmetric.
    if a.cholesky is None:
        a.cholesky = B.matmul(a.left, B.cholesky(matrix(a.middle)))
    return a.cholesky


# Matrix multiplications:

@B.matmul.extend(Dense, Dense, precedence=-1)
def matmul(a, b, tr_a=False, tr_b=False):
    # This is the fallback implementation. Any other specialised implementation
    # should be preferred over this.
    return B.matmul(dense(a), dense(b), tr_a=tr_a, tr_b=tr_b)


@B.matmul.extend(Diagonal, {B.Numeric, Dense})
def matmul(a, b, tr_a=False, tr_b=False):
    a = B.transpose(a) if tr_a else a
    b = B.transpose(b) if tr_b else b

    # Get shape of `a`.
    a_rows, a_cols = B.shape(a)

    # If `a` is square, don't do complicated things.
    if a_rows == a_cols and a_rows is not None:
        return B.diag(a)[:, None] * dense(b)

    # Compute the core part.
    rows = B.minimum(a_rows, B.shape(b)[0])
    core = B.diag(a)[:rows, None] * dense(b)[:rows, :]

    # Compute extra zeros to be appended.
    extra_rows = a_rows - rows
    extra_zeros = B.zeros(B.dtype(b), extra_rows, B.shape(b)[1])
    return B.concat(core, extra_zeros, axis=0)


@B.matmul.extend({B.Numeric, Dense}, Diagonal)
def matmul(a, b, tr_a=False, tr_b=False):
    a = B.transpose(a) if tr_a else a
    b = B.transpose(b) if tr_b else b

    # Get shape of `b`.
    b_rows, b_cols = B.shape(b)

    # If `b` is square, don't do complicated things.
    if b_rows == b_cols and b_rows is not None:
        return dense(a) * B.diag(b)[None, :]

    # Compute the core part.
    cols = B.minimum(B.shape(a)[1], b_cols)
    core = dense(a)[:, :cols] * B.diag(b)[None, :cols]

    # Compute extra zeros to be appended.
    extra_cols = b_cols - cols
    extra_zeros = B.zeros(B.dtype(b), B.shape(a)[0], extra_cols)
    return B.concat(core, extra_zeros, axis=1)


@B.matmul.extend(Diagonal, Diagonal)
def matmul(a, b, tr_a=False, tr_b=False):
    a = B.transpose(a) if tr_a else a
    b = B.transpose(b) if tr_b else b
    diag_len = B.minimum(B.diag_len(a), B.diag_len(b))
    return Diagonal(B.diag(a)[:diag_len] * B.diag(b)[:diag_len],
                    rows=B.shape(a)[0],
                    cols=B.shape(b)[1])


@B.matmul.extend(LowRank, LowRank)
def matmul(a, b, tr_a=False, tr_b=False):
    a = B.transpose(a) if tr_a else a
    b = B.transpose(b) if tr_b else b
    middle_inner = B.matmul(a.right, b.left, tr_a=True)
    middle = B.matmul(a.middle, middle_inner, b.middle)
    return LowRank(left=a.left, right=b.right, middle=middle)


@B.matmul.extend(LowRank, {B.Numeric, Dense})
def matmul(a, b, tr_a=False, tr_b=False):
    a = B.transpose(a) if tr_a else a
    b = B.transpose(b) if tr_b else b
    return LowRank(left=a.left,
                   right=B.matmul(b, a.right, tr_a=True),
                   middle=a.middle)


@B.matmul.extend({B.Numeric, Dense}, LowRank)
def matmul(a, b, tr_a=False, tr_b=False):
    a = B.transpose(a) if tr_a else a
    b = B.transpose(b) if tr_b else b
    return LowRank(left=B.matmul(a, b.left),
                   right=b.right,
                   middle=b.middle)


@B.matmul.extend(Diagonal, LowRank)
def matmul(a, b, tr_a=False, tr_b=False):
    # TODO: Refactor this to use ambiguous method resolution.
    return B.matmul.invoke(Dense, LowRank)(a, b, tr_a, tr_b)


@B.matmul.extend(LowRank, Diagonal)
def matmul(a, b, tr_a=False, tr_b=False):
    # TODO: Refactor this to use ambiguous method resolution.
    return B.matmul.invoke(LowRank, Dense)(a, b, tr_a, tr_b)


@B.matmul.extend({B.Numeric, Dense}, Woodbury, precedence=2)
def matmul(a, b, tr_a=False, tr_b=False):
    # Prioritise expanding out the Woodbury matrix. Give this one even higher
    # precedence to resolve ambiguity in the case of two Woodbury matrices.
    return B.add(B.matmul(a, b.lr, tr_a=tr_a, tr_b=tr_b),
                 B.matmul(a, b.diag, tr_a=tr_a, tr_b=tr_b))


@B.matmul.extend(Woodbury, {B.Numeric, Dense}, precedence=1)
def matmul(a, b, tr_a=False, tr_b=False):
    # Prioritise expanding out the Woodbury matrix.
    return B.add(B.matmul(a.lr, b, tr_a=tr_a, tr_b=tr_b),
                 B.matmul(a.diag, b, tr_a=tr_a, tr_b=tr_b))


@B.matmul.extend({B.Numeric, Dense},
                 {B.Numeric, Dense},
                 {B.Numeric, Dense},
                 [{B.Numeric, Dense}])
def matmul(*xs, **trs):
    def tr(name):
        return trs['tr_' + name] if 'tr_' + name in trs else False

    # Compute the first product.
    res = B.matmul(xs[0], xs[1], tr_a=tr('a'), tr_b=tr('b'))

    # Compute the remaining products.
    for name, x in zip(ascii_lowercase[2:], xs[2:]):
        res = B.matmul(res, x, tr_b=tr(name))

    return res


# Matrix inversion:

@_dispatch({B.Numeric, Dense})
def inverse(a):
    # Assume that `a` is PD.
    a = matrix(a)
    if a.inverse is None:
        inv_prod = B.trisolve(B.cholesky(a), B.eye(a))
        a.inverse = B.matmul(inv_prod, inv_prod, tr_a=True)
    return a.inverse


B.inverse = inverse  # Record in LAB.


@B.inverse.extend(Diagonal)
def inverse(a): return Diagonal(1 / B.diag(a), *reversed(B.shape(a)))


@B.inverse.extend(Woodbury)
def inverse(a):
    # Use the Woodbury matrix identity.
    if a.inverse is None:
        inv_diag = B.inverse(a.diag)
        lr = LowRank(left=B.matmul(inv_diag, a.lr.left),
                     right=B.matmul(inv_diag, a.lr.right),
                     middle=B.inverse(B.schur(a)))
        if a.lr_pd:
            a.inverse = Woodbury(diag=inv_diag, lr=-lr, lr_pd=False)
        else:
            a.inverse = Woodbury(diag=inv_diag, lr=lr, lr_pd=True)
    return a.inverse


def lr_diff(a, b):
    """Subtract two low-rank matrices, forcing the resulting middle part to
    be positive definite if the result is so.

    Args:
        a (:class:`.matrix.LowRank`): `a`.
        b (:class:`.matrix.LowRank`): `b`.

    Returns:
        :class:`.matrix.LowRank`: Difference between `a` and `b`.
    """
    diff = a - b
    u_left, s_left, v_left = B.svd(diff.left)
    u_right, s_right, v_right = B.svd(diff.right)
    middle = B.matmul(Diagonal(s_left, *B.shape(diff.left)),
                      B.matmul(v_left, diff.middle, v_right, tr_a=True),
                      Diagonal(s_right, *B.shape(diff.right)).T)
    return LowRank(left=u_left,
                   right=u_right,
                   middle=middle)


B.lr_diff = lr_diff  # Record in LAB.


@B.inverse.extend(LowRank)
def inverse(a): raise RuntimeError('Matrix is singular.')


# Compute the log-determinant of matrices.

@_dispatch({B.Numeric, Dense})
def logdet(a):
    a = matrix(a)
    if a.logdet is None:
        a.logdet = 2 * B.sum(B.log(B.diag(B.cholesky(a))))
    return a.logdet


B.logdet = logdet  # Record in LAB.


@B.logdet.extend(Diagonal)
def logdet(a): return B.sum(B.log(a.diag))


@B.logdet.extend(Woodbury)
def logdet(a):
    return B.logdet(B.schur(a)) + \
           B.logdet(a.diag) + \
           B.logdet(a.lr.middle if a.lr_pd else -a.lr.middle)


@B.logdet.extend(LowRank)
def logdet(a): raise RuntimeError('Matrix is singular.')


# Compute roots of matrices.

@_dispatch({B.Numeric, Dense})
def root(a):
    # TODO: This assumed that `a` is PSD.
    a = matrix(a)
    if a.root is None:
        u, s, _ = B.svd(dense(a), compute_uv=True)
        a.root = B.matmul(u * s[None, :] ** .5, u, tr_b=True)
    return a.root


B.root = root  # Record in LAB.


@B.root.extend(Diagonal)
def root(a):
    return Diagonal(a.diag ** .5)


# Sample from covariance matrices.

@_dispatch({B.Numeric, Dense}, [object])
def sample(a, num=1):
    """Sample from covariance matrices.

    Args:
        a (tensor): Covariance matrix to sample from.
        num (int): Number of samples.

    Returns:
        tensor: Samples as rank 2 column vectors.
    """
    # # Convert integer data types to floats.
    dtype = float if B.issubdtype(B.dtype(a), np.integer) else B.dtype(a)

    # Perform sampling operation.
    chol = B.cholesky(matrix(a))
    return B.matmul(chol, B.randn(dtype, B.shape(chol)[1], num))


B.sample = sample  # Record in LAB.


@B.sample.extend(Woodbury, [object])
def sample(a, num=1): return B.sample(a.diag, num) + B.sample(a.lr, num)


# Compute Schur complements.

@_dispatch(object, object, object, object)
def schur(a, b, c, d):
    """Compute the Schur complement `a - transpose(b) inv(c) d`.

    Args:
        a (tensor): `a`.
        b (tensor): `b`.
        c (tensor): `c`.
        d (tensor): `d`.

    Returns:
        :class:`.matrix.Dense`: Schur complement.
    """
    return B.subtract(a, B.qf(c, b, d))


B.schur = schur  # Record in LAB.


@B.schur.extend(object, object, Woodbury, object)
def schur(a, b, c, d):
    # The subtraction will yield a negative-definite Woodbury matrix. Convert
    # to dense to avoid issues with positive definiteness later on. This is
    # suboptimal, but more complicated abstractions are required to exploit
    # this.
    return dense(B.subtract(a, B.qf(c, b, d)))


@B.schur.extend(Woodbury)
def schur(a):
    if a.schur is None:
        prod = B.matmul(a.lr.right, B.inverse(a.diag), a.lr.left, tr_a=True)
        if a.lr_pd:
            a.schur = B.inverse(a.lr.middle) + prod
        else:
            a.schur = B.inverse(-a.lr.middle) - prod
    return a.schur


# Compute quadratic forms and diagonals thereof.

@_dispatch(object, object)
def qf(a, b):
    """Compute the quadratic form `transpose(b) inv(a) c`.

    Args:
        a (tensor): Covariance matrix.
        b (tensor): `b`.
        c (tensor, optional): `c`. Defaults to `b`.

    Returns:
        :class:`.matrix.Dense`: Quadratic form.
    """
    prod = B.trisolve(B.cholesky(matrix(a)), b)
    return matrix(B.matmul(prod, prod, tr_a=True))


B.qf = qf  # Record in LAB.


@B.qf.extend(object, object, object)
def qf(a, b, c):
    if b is c:
        return B.qf(a, b)
    chol = B.cholesky(matrix(a))
    return B.matmul(B.trisolve(chol, b), B.trisolve(chol, c), tr_a=True)


@B.qf.extend({Diagonal, Woodbury}, object)
def qf(a, b): return B.qf(a, b, b)


@B.qf.extend({Diagonal, Woodbury}, object, object)
def qf(a, b, c): return B.matmul(b, B.inverse(a), c, tr_a=True)


@B.qf.extend_multi((LowRank, object), (LowRank, object, object))
def qf(a, b, c=None): raise RuntimeError('Matrix is singular.')


@_dispatch(object, object)
def qf_diag(a, b):
    """Compute the diagonal of `transpose(b) inv(a) c`.

    Args:
        a (:class:`.matrix.Dense`): Covariance matrix.
        b (tensor): `b`.
        c (tensor, optional): `c`. Defaults to `b`.

    Returns:
        tensor: Diagonal of the quadratic form.
    """
    prod = B.trisolve(B.cholesky(matrix(a)), b)
    return B.sum(prod ** 2, axis=0)


B.qf_diag = qf_diag  # Record in LAB.


@B.qf_diag.extend(object, object, object)
def qf_diag(a, b, c):
    a = matrix(a)
    if b is c:
        return B.qf_diag(a, b)
    left = B.trisolve(B.cholesky(a), b)
    right = B.trisolve(B.cholesky(a), c)
    return B.sum(left * right, axis=0)


@B.qf_diag.extend({Diagonal, Woodbury}, object)
def qf_diag(a, b): return B.qf_diag(a, b, b)


@B.qf_diag.extend(Woodbury, object, object)
def qf_diag(a, b, c): return B.diag(B.qf(a, b, c))


@B.qf_diag.extend(Diagonal, object, object)
def qf_diag(a, b, c): return B.sum(B.matmul(B.inverse(a), b) * c, axis=0)


@B.qf_diag.extend_multi((LowRank, object), (LowRank, object, object))
def qf_diag(a, b, c=None): raise RuntimeError('Matrix is singular.')


# Compute ratios between matrices.

@_dispatch(object, object)
def ratio(a, b):
    """Compute the ratio between two positive-definite matrices.

    Args:
        a (tensor): Numerator.
        b (tensor): Denominator.

    Returns:
        tensor: Ratio.
    """
    return B.sum(B.qf_diag(b, B.cholesky(matrix(a))))


B.ratio = ratio  # Record in LAB.


@B.ratio.extend(LowRank, Dense)
def ratio(a, b): return B.sum(B.qf_diag(b, a.right, B.matmul(a.left, a.middle)))


@B.ratio.extend(Diagonal, Diagonal)
def ratio(a, b): return B.sum(a.diag / b.diag)


# Transpose matrices.

@B.transpose.extend(Dense)
def transpose(a): return matrix(B.transpose(dense(a)))


@B.transpose.extend(Diagonal)
def transpose(a): return Diagonal(B.diag(a), *reversed(B.shape(a)))


@B.transpose.extend(LowRank)
def transpose(a): return LowRank(left=a.right,
                                 right=a.left,
                                 middle=B.transpose(a.middle))


@B.transpose.extend(Constant)
def transpose(a): return Constant(a.constant, *reversed(B.shape(a)))


@B.transpose.extend(Woodbury)
def transpose(a): return Woodbury(diag=B.transpose(a.diag),
                                  lr=B.transpose(a.lr),
                                  lr_pd=a.lr_pd)


# Extend LAB to the field of matrices.

B.add.extend(Dense, object)(add)
B.add.extend(object, Dense)(add)
B.add.extend(Dense, Dense)(add)

B.multiply.extend(Dense, object)(mul)
B.multiply.extend(object, Dense)(mul)
B.multiply.extend(Dense, Dense)(mul)


@B.subtract.extend(Dense, Dense)
def subtract(a, b): return B.add(a, -b)


@B.divide.extend(Dense, Dense)
def divide(a, b): return B.multiply(a, 1 / b)


# Get shapes of matrices.


@B.shape.extend(Dense)
def shape(a): return B.shape(dense(a))


#   Don't use a union here to prevent ambiguity errors with the implementation
#     for `LowRank`.
@B.shape.extend_multi((Diagonal,), (Constant,))
def shape(a): return a.rows, a.cols


@B.shape.extend(LowRank)
def shape(a): return B.shape(a.left)[0], B.shape(a.right)[0]


@B.shape.extend(Woodbury)
def shape(a): return B.shape(a.lr)


# Setup promotion and conversion of matrices as a fallback mechanism.

add_promotion_rule(B.Numeric, Dense, B.Numeric)
add_conversion_method(Dense, B.Numeric, dense)


# Simplify addiction and multiplication between matrices.

@mul.extend(Dense, Dense)
def mul(a, b): return Dense(dense(a) * dense(b))


@mul.extend(Dense, Diagonal)
def mul(a, b): return Diagonal(B.diag(a) * b.diag, *B.shape(a))


@mul.extend(Diagonal, Dense)
def mul(a, b): return Diagonal(a.diag * B.diag(b), *B.shape(a))


@mul.extend(Diagonal, Diagonal)
def mul(a, b): return Diagonal(a.diag * b.diag, *B.shape(a))


@mul.extend(LowRank, LowRank)
def mul(a, b):
    # Pick apart the matrices.
    al, ar = B.unstack(a.left, axis=1), B.unstack(a.right, axis=1)
    bl, br = B.unstack(b.left, axis=1), B.unstack(b.right, axis=1)
    am = [B.unstack(x, axis=0) for x in B.unstack(a.middle, axis=0)]
    bm = [B.unstack(x, axis=0) for x in B.unstack(b.middle, axis=0)]

    # Construct all parts of the product.
    left = B.stack(*[ali * blk
                     for ali in al
                     for blk in bl], axis=1)
    right = B.stack(*[arj * brl
                      for arj in ar
                      for brl in br], axis=1)
    middle = B.stack(*[B.stack(*[amij * bmkl
                                 for amij in ami
                                 for bmkl in bmk], axis=0)
                       for ami in am
                       for bmk in bm], axis=0)

    # And assemble the result.
    return LowRank(left=left, right=right, middle=middle)


@mul.extend(Dense, Woodbury, precedence=2)
def mul(a, b):
    # Prioritise expanding out the Woodbury matrix. Give this one even higher
    # precedence to resolve ambiguity in the case of two Woodbury matrices.
    return add(mul(a, b.lr), mul(a, b.diag))


@mul.extend(Woodbury, Dense, precedence=1)
def mul(a, b):
    # Prioritise expanding out the Woodbury matrix.
    return add(mul(a.lr, b), mul(a.diag, b))


@add.extend(Dense, Dense)
def add(a, b): return Dense(dense(a) + dense(b))


@add.extend(Diagonal, Diagonal)
def add(a, b): return Diagonal(B.diag(a) + B.diag(b), *B.shape(a))


@add.extend(LowRank, LowRank)
def add(a, b):
    shape_a, shape_b, dtype = B.shape(a.middle), B.shape(b.middle), B.dtype(a)
    middle = B.concat2d([a.middle, B.zeros(dtype, shape_a[0], shape_b[1])],
                        [B.zeros(dtype, shape_b[0], shape_a[1]), b.middle])
    return LowRank(left=B.concat(a.left, b.left, axis=1),
                   right=B.concat(a.right, b.right, axis=1),
                   middle=middle)


# Woodbury matrices:

@add.extend(LowRank, Diagonal)
def add(a, b): return Woodbury(diag=b, lr=a)


@add.extend(Diagonal, LowRank)
def add(a, b): return Woodbury(diag=a, lr=b)


@add.extend(LowRank, Woodbury)
def add(a, b): return Woodbury(diag=b.diag, lr=a + b.lr)


@add.extend(Woodbury, LowRank)
def add(a, b): return Woodbury(diag=a.diag, lr=a.lr + b)


@add.extend(Diagonal, Woodbury)
def add(a, b): return Woodbury(diag=a + b.diag, lr=b.lr)


@add.extend(Woodbury, Diagonal)
def add(a, b): return Woodbury(diag=a.diag + b, lr=a.lr)


@add.extend(Woodbury, Woodbury)
def add(a, b): return Woodbury(diag=a.diag + b.diag, lr=a.lr + b.lr)


# Multiplication between matrices and constants.

@mul.extend(LowRank, Constant)
def mul(a, b): return LowRank(left=a.left,
                              right=a.right,
                              middle=a.middle * b.constant)


@mul.extend(Constant, LowRank)
def mul(a, b): return LowRank(left=b.left,
                              right=b.right,
                              middle=a.constant * b.middle)


@mul.extend(Diagonal, Constant)
def mul(a, b): return Diagonal(B.diag(a) * b.constant, *B.shape(a))


@mul.extend(Constant, Diagonal)
def mul(a, b): return Diagonal(a.constant * B.diag(b), *B.shape(a))


@mul.extend(Constant, Constant)
def mul(a, b): return Constant(a.constant * b.constant, *B.shape(a))


# Simplify addiction and multiplication between matrices and other objects. We
# immediately resolve scaled elements.

@mul.extend(object, Dense)
def mul(a, b): return mul(b, a)


@mul.extend(Dense, object)
def mul(a, b): return mul(a, mul(One.from_(a), b))


@add.extend(object, Dense)
def add(a, b): return add(b, a)


@add.extend(Dense, object)
def add(a, b): return add(a, mul(One.from_(a), b))


@mul.extend(object, One)
def mul(a, b): return mul(b, a)


@mul.extend(One, object)
def mul(a, b):
    if B.rank(b) == 0:
        if b is 0:
            return Zero.from_(a)
        elif b is 1:
            return a
        else:
            return Constant.from_(b, a)
    else:
        return matrix(dense(a) * b)


# Take over construction of ones: construction requires arguments.

@add.extend(object, Zero)
def add(a, b): return add(b, a)


@add.extend(Zero, object)
def add(a, b): return mul(b, One.from_(a))


@add.extend(Zero, Zero)
def add(a, b): return a



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from lab import B
from plum import Dispatcher, Referentiable, Self, NotFoundLookupError

from .util import uprank

__all__ = []

_dispatch = Dispatcher()

Formatter = object  #: A formatter can be any object.
priority = 10  #: Priority precedence level.
definite = 20  #: Highest precedence level.


def squeeze(xs):
    """Squeeze an sequence if it only contains a single element.

    Args:
        xs (sequence): Sequence to squeeze.

    Returns:
        object: `xs[0]` if `xs` consists of a single element and `xs` otherwise.
    """
    return xs[0] if len(xs) == 1 else xs


def get_subclasses(c):
    """Get all subclasses of a class.

    Args:
        c (type): Class to get subclasses of.

    Returns:
        list[type]: List of subclasses of `c`.
    """
    return c.__subclasses__() + \
           [x for sc in c.__subclasses__() for x in get_subclasses(sc)]


def broadcast(op, xs, ys):
    """Perform a binary operation `op` on elements of `xs` and `ys`. If `xs` or
    `ys` has length 1, then it is repeated sufficiently many times to match the
    length of the other.

    Args:
        op (function): Binary operation.
        xs (sequence): First sequence.
        ys (sequence): Second sequence.

    Returns:
        tuple: Result of applying `op` to every element of `zip(xs, ys)` after
        broadcasting appropriately.
    """
    if len(xs) == 1 and len(ys) > 1:
        # Broadcast `xs`.
        xs = xs * len(ys)
    elif len(ys) == 1 and len(xs) > 1:
        # Broadcast `ys.
        ys = ys * len(xs)

    # Check that `xs` and `ys` are compatible now.
    if len(xs) != len(ys):
        raise ValueError('Inputs "{}" and "{}" could not be broadcasted.'
                         ''.format(xs, ys))
    # Perform operation.
    return tuple(op(x, y) for x, y in zip(xs, ys))


class Element(Referentiable):
    """A field over functions.

    Functions are also referred to as elements of the field. Elements can be
    added and multiplied.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __eq__(self, other):
        return False

    def __mul__(self, other):
        return mul(self, other)

    def __rmul__(self, other):
        return mul(other, self)

    def __add__(self, other):
        return add(self, other)

    def __radd__(self, other):
        return add(other, self)

    def __neg__(self):
        return mul(-1, self)

    def __sub__(self, other):
        return add(self, -other)

    def __rsub__(self, other):
        return add(other, -self)

    @property
    def num_terms(self):
        """Number of terms"""
        return 1

    def term(self, i):
        """Get a specific term.

        Args:
            i (int): Index of term.

        Returns:
            :class:`.field.Element`: The referenced term.
        """
        if i == 0:
            return self
        else:
            raise IndexError('Index out of range.')

    @property
    def num_factors(self):
        """Number of factors"""
        return 1

    def factor(self, i):
        """Get a specific factor.

        Args:
            i (int): Index of factor.

        Returns:
            :class:`.field.Element`: The referenced factor.
        """
        if i == 0:
            return self
        else:
            raise IndexError('Index out of range.')

    @property
    def __name__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.display()

    def __str__(self):
        return self.display()

    @_dispatch(Formatter)
    def display(self, formatter):
        """Display the element.

        Args:
            formatter (function, optional): Function to format values.

        Returns:
            str: Element as a string.
        """
        # Due to multiple inheritance, we might arrive here before arriving at a
        # method in the appropriate subclass. The only case to consider is if
        # we're not in a leaf.
        if isinstance(self, (JoinElement, WrappedElement)):
            return pretty_print(self, formatter)
        else:
            return self.__class__.__name__ + '()'

    @_dispatch()
    def display(self):
        return self.display(lambda x: x)


class OneElement(Element, Referentiable):
    """The constant `1`."""
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(Formatter)
    def display(self, formatter):
        return '1'

    @_dispatch(Self)
    def __eq__(self, other):
        return True


class ZeroElement(Element, Referentiable):
    """The constant `0`."""
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(Formatter)
    def display(self, formatter):
        return '0'

    @_dispatch(Self)
    def __eq__(self, other):
        return True


class WrappedElement(Element, Referentiable):
    """A wrapped element.

    Args:
        e (:class:`.field.Element`): Element to wrap.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e):
        self.e = e

    def __getitem__(self, item):
        if item == 0:
            return self.e
        else:
            raise IndexError('Index out of range.')

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        raise NotImplementedError()


class JoinElement(Element, Referentiable):
    """Two wrapped elements.

    Args:
        e1 (:class:`.field.Element`): First element to wrap.
        e2 (:class:`.field.Element`): Second element to wrap.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e1, e2):
        self.e1 = e1
        self.e2 = e2

    def __getitem__(self, item):
        if item == 0:
            return self.e1
        elif item == 1:
            return self.e2
        else:
            raise IndexError('Index out of range.')

    @_dispatch(object, object, Formatter)
    def display(self, e1, e2, formatter):
        raise NotImplementedError()


class ScaledElement(WrappedElement, Referentiable):
    """Scaled element.

    Args:
        e (:class:`.field.Element`): Element to scale.
        scale (tensor): Scale.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e, scale):
        WrappedElement.__init__(self, e)
        self.scale = scale

    @property
    def num_factors(self):
        return self[0].num_factors + 1

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        return '{} * {}'.format(formatter(self.scale), e)

    def factor(self, i):
        if i >= self.num_factors:
            raise IndexError('Index out of range.')
        else:
            return self.scale if i == 0 else self[0].factor(i - 1)

    @_dispatch(Self)
    def __eq__(self, other):
        return B.all(self.scale == other.scale) and self[0] == other[0]


class ProductElement(JoinElement, Referentiable):
    """Product of elements."""
    _dispatch = Dispatcher(in_class=Self)

    @property
    def num_factors(self):
        return self[0].num_factors + self[1].num_factors

    def factor(self, i):
        if i >= self.num_factors:
            raise IndexError('Index out of range.')
        if i < self[0].num_factors:
            return self[0].factor(i)
        else:
            return self[1].factor(i - self[0].num_factors)

    @_dispatch(object, object, Formatter)
    def display(self, e1, e2, formatter):
        return '{} * {}'.format(e1, e2)

    @_dispatch(Self)
    def __eq__(self, other):
        return (self[0] == other[0] and self[1] == other[1]) or \
               (self[0] == other[1] and self[1] == other[0])


class SumElement(JoinElement, Referentiable):
    """Sum of elements."""
    _dispatch = Dispatcher(in_class=Self)

    @property
    def num_terms(self):
        return self[0].num_terms + self[1].num_terms

    def term(self, i):
        if i >= self.num_terms:
            raise IndexError('Index out of range.')
        if i < self[0].num_terms:
            return self[0].term(i)
        else:
            return self[1].term(i - self[0].num_terms)

    @_dispatch(object, object, Formatter)
    def display(self, e1, e2, formatter):
        return '{} + {}'.format(e1, e2)

    @_dispatch(Self)
    def __eq__(self, other):
        return (self[0] == other[0] and self[1] == other[1]) or \
               (self[0] == other[1] and self[1] == other[0])


@_dispatch(object, object)
def mul(a, b):
    """Multiply two elements.

    Args:
        a (:class:`.field.Element`): First element in product.
        b (:class:`.field.Element`): Second element in product.

    Returns:
        :class:`.field.Element`: Product of the elements.
    """
    raise NotImplementedError('Multiplication not implemented for "{}" and '
                              '"{}"'.format(type(a).__name__, type(b).__name__))


@_dispatch(object, object)
def add(a, b):
    """Add two elements.

    Args:
        a (:class:`.field.Element`): First element in addition.
        b (:class:`.field.Element`): Second element in addition.

    Returns:
        :class:`.field.Element`: Sum of the elements.
    """
    raise NotImplementedError('Addition not implemented for "{}" and "{}".'
                              ''.format(type(a).__name__, type(b).__name__))


@_dispatch(object)
def get_field(a):
    """Get the field of an element.

    Args:
        a (:class:`.field.Element`): Element to get field of.

    Returns:
        type: Field of `a`.
    """
    raise RuntimeError('Could not determine field type of "{}".'
                       ''.format(type(a).__name__))


new_cache = {}


def new(a, t):
    """Create a new specialised type.

    Args:
        a (:class:`.field.Element`): Element to create new type for.
        t (type): Type to create.

    Returns:
        type: Specialisation of `t` appropriate for `a`.
    """
    try:
        return new_cache[type(a), t]
    except KeyError:
        field = get_field(a)
        candidates = list(set(get_subclasses(field)) & set(get_subclasses(t)))

        # There should only be a single candidate.
        if len(candidates) != 1:
            raise RuntimeError('Could not determine "{}" for field "{}".'
                               ''.format(t.__name__, field.__name__))

        new_cache[type(a), t] = candidates[0]
        return new_cache[type(a), t]


# Pretty printing with minimal parentheses.

@_dispatch(Element, Formatter)
def pretty_print(el, formatter):
    """Pretty print an element with a minimal number of parentheses.

    Args:
        el (:class:`.field.Element`): Element to print.
        formatter (:class:`.field.Formatter`): Formatter for values.

    Returns:
        str: `el` converted to string prettily.
    """
    return el.display(formatter)


@_dispatch(WrappedElement, Formatter)
def pretty_print(el, formatter):
    return el.display(pretty_print(el[0], el, formatter), formatter)


@_dispatch(JoinElement, Formatter)
def pretty_print(el, formatter):
    return el.display(pretty_print(el[0], el, formatter),
                      pretty_print(el[1], el, formatter), formatter)


@_dispatch(Element, Element, Formatter)
def pretty_print(el, parent, formatter):
    if need_parens(el, parent):
        return '(' + pretty_print(el, formatter) + ')'
    else:
        return pretty_print(el, formatter)


@_dispatch(Element, SumElement)
def need_parens(el, parent):
    """Check whether `el` needs parentheses when printed in `parent`.

    Args:
        el (:class:`.field.Element`): Element to print.
        parent (:class:`.field.Element`): Parent of element to print.

    Returns:
        bool: Boolean whether `el` needs parentheses.
    """
    return False


@_dispatch(Element, ProductElement)
def need_parens(el, parent): return False


@_dispatch({SumElement, WrappedElement}, ProductElement)
def need_parens(el, parent): return True


@_dispatch(ScaledElement, ProductElement)
def need_parens(el, parent): return False


@_dispatch(Element, WrappedElement)
def need_parens(el, parent): return False


@_dispatch({WrappedElement, JoinElement}, WrappedElement)
def need_parens(el, parent): return True


@_dispatch({ProductElement, ScaledElement}, ScaledElement)
def need_parens(el, parent): return False


# Generic multiplication.

@_dispatch(Element, object)
def mul(a, b):
    if b is 0:
        return new(a, ZeroElement)()
    elif b is 1:
        return a
    else:
        return new(a, ScaledElement)(a, b)


@_dispatch(object, Element)
def mul(a, b):
    return mul(b, a)


@_dispatch(Element, Element)
def mul(a, b): return new(a, ProductElement)(a, b)


# Generic addition.

@_dispatch(Element, object)
def add(a, b):
    if b is 0:
        return a
    else:
        return new(a, SumElement)(a, mul(b, new(a, OneElement)()))


@_dispatch(object, Element)
def add(a, b):
    if a is 0:
        return b
    else:
        return new(b, SumElement)(mul(a, new(b, OneElement)()), b)


@_dispatch(Element, Element)
def add(a, b):
    if a == b:
        return mul(2, a)
    else:
        return new(a, SumElement)(a, b)


# Cancel redundant zeros and ones.

@_dispatch(ZeroElement, object, precedence=definite)
def mul(a, b): return a


@_dispatch(object, ZeroElement, precedence=definite)
def mul(a, b): return b


@_dispatch(ZeroElement, ZeroElement, precedence=definite)
def mul(a, b): return a


@_dispatch(OneElement, Element, precedence=priority)
def mul(a, b): return b


@_dispatch(Element, OneElement, precedence=priority)
def mul(a, b): return a


@_dispatch(OneElement, OneElement, precedence=priority)
def mul(a, b): return a


@_dispatch(ZeroElement, object, precedence=definite)
def add(a, b):
    if b is 0:
        return a
    else:
        return mul(new(a, OneElement)(), b)


@_dispatch(object, ZeroElement, precedence=definite)
def add(a, b):
    if a is 0:
        return b
    else:
        return mul(a, new(b, OneElement)())


@_dispatch(ZeroElement, ZeroElement, precedence=definite)
def add(a, b): return a


@_dispatch(Element, ZeroElement, precedence=definite)
def add(a, b): return a


@_dispatch(ZeroElement, Element, precedence=definite)
def add(a, b): return b


# Group factors and terms if possible.

@_dispatch(object, ScaledElement)
def mul(a, b): return mul(b.scale * a, b[0])


@_dispatch(ScaledElement, object)
def mul(a, b): return mul(a.scale * b, a[0])


@_dispatch(ScaledElement, Element)
def mul(a, b): return mul(a.scale, mul(a[0], b))


@_dispatch(Element, ScaledElement)
def mul(a, b): return mul(b.scale, mul(a, b[0]))


@_dispatch(ScaledElement, ScaledElement)
def mul(a, b):
    if a[0] == b[0]:
        return new(a, ScaledElement)(a[0], a.scale * b.scale)
    else:
        scaled = new(a, ScaledElement)(a[0], a.scale * b.scale)
        return new(a, ProductElement)(scaled, b[0])


@_dispatch(ScaledElement, Element)
def add(a, b):
    if a[0] == b:
        return mul(a.scale + 1, b)
    else:
        return new(a, SumElement)(a, b)


@_dispatch(Element, ScaledElement)
def add(a, b):
    if a == b[0]:
        return mul(b.scale + 1, a)
    else:
        return new(a, SumElement)(a, b)


@_dispatch(ScaledElement, ScaledElement)
def add(a, b):
    if a[0] == b[0]:
        return mul(a.scale + b.scale, a[0])
    else:
        return new(a, SumElement)(a, b)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import numpy as np
import pytest
from lab import B

from stheno.graph import GP, Graph
from stheno.input import Component
from stheno.kernel import OneKernel, EQ
from stheno.mean import OneMean
from stheno.util import uprank
from .util import allclose, approx


def test_uprank():
    allclose(uprank(0), [[0]])
    allclose(uprank(np.array([0])), [[0]])
    allclose(uprank(np.array([[0]])), [[0]])
    assert type(uprank(Component('test')(0))) == Component('test')

    k = OneKernel()

    assert B.shape(k(0, 0)) == (1, 1)
    assert B.shape(k(0, np.ones(5))) == (1, 5)
    assert B.shape(k(0, np.ones((5, 2)))) == (1, 5)

    assert B.shape(k(np.ones(5), 0)) == (5, 1)
    assert B.shape(k(np.ones(5), np.ones(5))) == (5, 5)
    assert B.shape(k(np.ones(5), np.ones((5, 2)))) == (5, 5)

    assert B.shape(k(np.ones((5, 2)), 0)) == (5, 1)
    assert B.shape(k(np.ones((5, 2)), np.ones(5))) == (5, 5)
    assert B.shape(k(np.ones((5, 2)), np.ones((5, 2)))) == (5, 5)

    with pytest.raises(ValueError):
        k(0, np.ones((5, 2, 1)))
    with pytest.raises(ValueError):
        k(np.ones((5, 2, 1)))

    m = OneMean()

    assert B.shape(m(0)) == (1, 1)
    assert B.shape(m(np.ones(5))) == (5, 1)
    assert B.shape(m(np.ones((5, 2)))) == (5, 1)

    p = GP(EQ(), graph=Graph())
    x = np.linspace(0, 10, 10)

    approx(p.condition(1, 1)(1).mean, np.array([[1]]))
    approx(p.condition(x, x)(x).mean, x[:, None])
    approx(p.condition(x, x[:, None])(x).mean, x[:, None])



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from itertools import product

import numpy as np
import pytest
from lab import B
from stheno.matrix import (
    Dense,
    Diagonal,
    LowRank,
    Woodbury,
    matrix,
    Zero,
    One,
    Constant
)

from .util import allclose, to_np


def test_equality():
    # Test `Dense.`
    a = Dense(np.random.randn(4, 2))
    allclose(a == a, to_np(a) == to_np(a))

    # Test `Diagonal`.
    d = Diagonal(np.random.randn(4))
    allclose(d == d, B.diag(d) == B.diag(d))

    # Test `LowRank`.
    lr = LowRank(left=np.random.randn(4, 2), middle=np.random.randn(2, 2))
    allclose(lr == lr, (lr.l == lr.l, lr.m == lr.m, lr.r == lr.r))

    # Test `Woodbury`.
    allclose((lr + d) == (lr + d), (lr == lr, d == d))

    # Test `Constant`.
    c1 = Constant.from_(1, a)
    c1_2 = Constant(1, 4, 3)
    c2 = Constant.from_(2, a)
    assert c1 == c1
    assert c1 != c1_2
    assert c1 != c2

    # Test `One`.
    one1 = One(np.float64, 4, 2)
    one2 = One(np.float64, 4, 3)
    assert one1 == one1
    assert one1 != one2

    # Test `Zero`.
    zero1 = Zero(np.float64, 4, 2)
    zero2 = Zero(np.float64, 4, 3)
    assert zero1 == zero1
    assert zero1 != zero2


def test_constant_zero_one():
    a = np.random.randn(4, 2)
    b = Dense(a)

    allclose(Zero.from_(a), np.zeros((4, 2)))
    allclose(Zero.from_(b), np.zeros((4, 2)))

    allclose(One.from_(a), np.ones((4, 2)))
    allclose(One.from_(b), np.ones((4, 2)))

    allclose(Constant.from_(2, a), 2 * np.ones((4, 2)))
    allclose(Constant.from_(2, b), 2 * np.ones((4, 2)))


def test_shorthands():
    a = Dense(np.random.randn(4, 4))
    allclose(a.T, B.transpose(a))
    allclose(a.__matmul__(a), B.matmul(a, a))


def test_dense_methods():
    a = Dense(np.random.randn(10, 5))
    allclose(a.T, to_np(a).T)
    allclose(-a, -to_np(a))
    allclose(a[5], to_np(a)[5])


def test_eye():
    a = Dense(np.random.randn(5, 10))
    allclose(B.eye(a), np.eye(5, 10))
    allclose(B.eye(a.T), np.eye(10, 5))


def test_matrix():
    a = np.random.randn(5, 3)
    assert type(matrix(a)) == Dense
    allclose(matrix(a).mat, a)
    assert type(matrix(matrix(a))) == Dense
    allclose(matrix(matrix(a)).mat, a)


def test_dense():
    a = np.random.randn(5, 3)
    allclose(Dense(a), a)

    # Extensively test Diagonal.
    allclose(Diagonal(np.array([1, 2])), np.array([[1, 0],
                                                   [0, 2]]))
    allclose(Diagonal(np.array([1, 2]), 3), np.array([[1, 0, 0],
                                                      [0, 2, 0],
                                                      [0, 0, 0]]))
    allclose(Diagonal(np.array([1, 2]), 1), np.array([[1]]))
    allclose(Diagonal(np.array([1, 2]), 3, 3), np.array([[1, 0, 0],
                                                         [0, 2, 0],
                                                         [0, 0, 0]]))
    allclose(Diagonal(np.array([1, 2]), 2, 3), np.array([[1, 0, 0],
                                                         [0, 2, 0]]))
    allclose(Diagonal(np.array([1, 2]), 3, 2), np.array([[1, 0],
                                                         [0, 2],
                                                         [0, 0]]))
    allclose(Diagonal(np.array([1, 2]), 1, 3), np.array([[1, 0, 0]]))
    allclose(Diagonal(np.array([1, 2]), 3, 1), np.array([[1],
                                                         [0],
                                                         [0]]))

    # Test low-rank matrices.
    left = np.random.randn(5, 3)
    right = np.random.randn(10, 3)
    middle = np.random.randn(3, 3)
    lr = LowRank(left=left, right=right, middle=middle)
    allclose(lr, left.dot(middle).dot(right.T))

    # Test Woodbury matrices.
    diag = Diagonal(np.array([1, 2, 3, 4]), 5, 10)
    wb = Woodbury(diag=diag, lr=lr)
    allclose(wb, to_np(diag) + to_np(lr))


def test_block_matrix():
    dt = np.float64

    # Check correctness.
    rows = [[np.random.randn(4, 3), np.random.randn(4, 5)],
            [np.random.randn(6, 3), np.random.randn(6, 5)]]
    allclose(B.block_matrix(*rows), B.concat2d(*rows))

    # Check that grid is checked correctly.
    assert type(B.block_matrix([Zero(dt, 3, 7), Zero(dt, 3, 4)],
                               [Zero(dt, 4, 5), Zero(dt, 4, 6)])) == Dense
    with pytest.raises(ValueError):
        B.block_matrix([Zero(dt, 5, 5), Zero(dt, 3, 6)],
                       [Zero(dt, 2, 5), Zero(dt, 4, 6)])

    # Test zeros.
    res = B.block_matrix([Zero(dt, 3, 5), Zero(dt, 3, 6)],
                         [Zero(dt, 4, 5), Zero(dt, 4, 6)])
    assert type(res) == Zero
    allclose(res, Zero(dt, 7, 11))

    # Test ones.
    res = B.block_matrix([One(dt, 3, 5), One(dt, 3, 6)],
                         [One(dt, 4, 5), One(dt, 4, 6)])
    assert type(res) == One
    allclose(res, One(dt, 7, 11))

    # Test diagonal.
    res = B.block_matrix([Diagonal(np.array([1, 2])), Zero(dt, 2, 3)],
                         [Zero(dt, 3, 2), Diagonal(np.array([3, 4, 5]))])
    assert type(res) == Diagonal
    allclose(res, Diagonal(np.array([1, 2, 3, 4, 5])))
    # Check that all blocks on the diagonal must be diagonal or zero.
    assert type(B.block_matrix([Diagonal(np.array([1, 2])), Zero(dt, 2, 3)],
                               [Zero(dt, 3, 2), One(dt, 3)])) == Dense
    assert type(B.block_matrix([Diagonal(np.array([1, 2])), Zero(dt, 2, 3)],
                               [Zero(dt, 3, 2), Zero(dt, 3)])) == Diagonal
    # Check that all blocks on the diagonal must be square.
    assert type(B.block_matrix([Diagonal(np.array([1, 2])), Zero(dt, 2, 4)],
                               [Zero(dt, 3, 2), Zero(dt, 3, 4)])) == Dense
    # Check that all other blocks must be zero.
    assert type(B.block_matrix([Diagonal(np.array([1, 2])), One(dt, 2, 3)],
                               [Zero(dt, 3, 2),
                                Diagonal(np.array([3, 4, 5]))])) == Dense


def test_dtype():
    # Test `Dense`.
    assert B.dtype(Dense(np.array([[1]]))) == np.int64
    assert B.dtype(Dense(np.array([[1.0]]))) == np.float64

    # Test `Diagonal`.
    diag_int = Diagonal(np.array([1]))
    diag_float = Diagonal(np.array([1.0]))
    assert B.dtype(diag_int) == np.int64
    assert B.dtype(diag_float) == np.float64

    # Test `LowRank`.
    lr_int = LowRank(left=np.array([[1]]),
                     right=np.array([[2]]),
                     middle=np.array([[3]]))
    lr_float = LowRank(left=np.array([[1.0]]),
                       right=np.array([[2.0]]),
                       middle=np.array([[3.0]]))
    assert B.dtype(lr_int) == np.int64
    assert B.dtype(lr_float) == np.float64

    # Test `Constant`.
    assert B.dtype(Constant(1, rows=1)) == int
    assert B.dtype(Constant(1.0, rows=1)) == float

    # Test `Woodbury`.
    assert B.dtype(Woodbury(diag_int, lr_int)) == np.int64
    assert B.dtype(Woodbury(diag_float, lr_float)) == np.float64


def test_sum():
    a = Dense(np.random.randn(10, 20))
    allclose(B.sum(a, axis=0), np.sum(to_np(a), axis=0))

    for x in [Diagonal(np.array([1, 2, 3]), rows=3, cols=5),
              LowRank(left=np.random.randn(5, 3),
                      right=np.random.randn(10, 3),
                      middle=np.random.randn(3, 3))]:
        allclose(B.sum(x), np.sum(to_np(x)))
        allclose(B.sum(x, axis=0), np.sum(to_np(x), axis=0))
        allclose(B.sum(x, axis=1), np.sum(to_np(x), axis=1))
        allclose(B.sum(x, axis=(0, 1)), np.sum(to_np(x), axis=(0, 1)))


def test_trace():
    a = np.random.randn(10, 10)
    allclose(B.trace(Dense(a)), np.trace(a))


def test_diag():
    # Test `Dense`.
    a = np.random.randn(5, 3)
    allclose(B.diag(Dense(a)), np.diag(a))

    # Test `Diagonal`.
    allclose(B.diag(Diagonal(np.array([1, 2, 3]))), [1, 2, 3])
    allclose(B.diag(Diagonal(np.array([1, 2, 3]), 2)), [1, 2])
    allclose(B.diag(Diagonal(np.array([1, 2, 3]), 4)), [1, 2, 3, 0])

    # Test `LowRank`.
    b = np.random.randn(10, 3)
    allclose(B.diag(LowRank(left=a, right=a)), np.diag(a.dot(a.T)))
    allclose(B.diag(LowRank(left=a, right=b)), np.diag(a.dot(b.T)))
    allclose(B.diag(LowRank(left=b, right=b)), np.diag(b.dot(b.T)))

    # Test `Constant`.
    allclose(B.diag(Constant(1, rows=3, cols=5)), np.ones(3))

    # Test `Woodbury`.
    allclose(B.diag(Woodbury(Diagonal(np.array([1, 2, 3]), rows=5, cols=10),
                             LowRank(left=a, right=b))),
             np.diag(a.dot(b.T) + np.concatenate((np.diag([1, 2, 3, 0, 0]),
                                                  np.zeros((5, 5))), axis=1)))

    # The method `B.diag(diag, rows, cols)` is tested in the tests for
    # `Diagonal` in `test_dense`.


def test_diag_len():
    assert B.diag_len(np.ones((5, 5))) == 5
    assert B.diag_len(np.ones((10, 5))) == 5
    assert B.diag_len(np.ones((5, 10))) == 5


def test_cholesky():
    a = np.random.randn(5, 5)
    a = a.T.dot(a)

    # Test `Dense`.
    allclose(np.linalg.cholesky(a), B.cholesky(a))

    # Test `Diagonal`.
    d = Diagonal(np.diag(a))
    allclose(np.linalg.cholesky(to_np(d)), B.cholesky(d))

    # Test `LowRank`.
    a = np.random.randn(2, 2)
    lr = LowRank(left=np.random.randn(5, 2), middle=a.dot(a.T))
    chol = to_np(B.cholesky(lr))
    #   The Cholesky here is not technically the Cholesky decomposition. Hence
    #   we test this slightly differently.
    allclose(chol.dot(chol.T), lr)


def test_matmul():
    diag_square = Diagonal(np.array([1, 2]), 3)
    diag_tall = Diagonal(np.array([3, 4]), 5, 3)
    diag_wide = Diagonal(np.array([5, 6]), 2, 3)

    dense_square = Dense(np.random.randn(3, 3))
    dense_tall = Dense(np.random.randn(5, 3))
    dense_wide = Dense(np.random.randn(2, 3))

    lr = LowRank(left=np.random.randn(5, 2),
                 right=np.random.randn(3, 2),
                 middle=np.random.randn(2, 2))

    def compare(a, b):
        return np.allclose(to_np(B.matmul(a, b)),
                           B.matmul(to_np(a), to_np(b)))

    # Test `Dense`.
    assert compare(dense_wide, dense_tall.T), 'dense w x dense t'

    # Test `LowRank`.
    assert compare(lr, dense_tall.T), 'lr x dense t'
    assert compare(dense_wide, lr.T), 'dense w x lr'
    assert compare(lr, diag_tall.T), 'lr x diag t'
    assert compare(diag_wide, lr.T), 'diag w x lr'
    assert compare(lr, lr.T), 'lr x lr'
    assert compare(lr.T, lr), 'lr x lr (2)'

    # Test `Diagonal`.
    #   Test multiplication between diagonal matrices.
    assert compare(diag_square, diag_square.T), 'diag s x diag s'
    assert compare(diag_tall, diag_square.T), 'diag t x diag s'
    assert compare(diag_wide, diag_square.T), 'diag w x diag s'
    assert compare(diag_square, diag_tall.T), 'diag s x diag t'
    assert compare(diag_tall, diag_tall.T), 'diag t x diag t'
    assert compare(diag_wide, diag_tall.T), 'diag w x diag t'
    assert compare(diag_square, diag_wide.T), 'diag s x diag w'
    assert compare(diag_tall, diag_wide.T), 'diag t x diag w'
    assert compare(diag_wide, diag_wide.T), 'diag w x diag w'

    #   Test multiplication between diagonal and dense matrices.
    assert compare(diag_square, dense_square.T), 'diag s x dense s'
    assert compare(diag_square, dense_tall.T), 'diag s x dense t'
    assert compare(diag_square, dense_wide.T), 'diag s x dense w'
    assert compare(diag_tall, dense_square.T), 'diag t x dense s'
    assert compare(diag_tall, dense_tall.T), 'diag t x dense t'
    assert compare(diag_tall, dense_wide.T), 'diag t x dense w'
    assert compare(diag_wide, dense_square.T), 'diag w x dense s'
    assert compare(diag_wide, dense_tall.T), 'diag w x dense t'
    assert compare(diag_wide, dense_wide.T), 'diag w x dense w'

    assert compare(dense_square, diag_square.T), 'dense s x diag s'
    assert compare(dense_square, diag_tall.T), 'dense s x diag t'
    assert compare(dense_square, diag_wide.T), 'dense s x diag w'
    assert compare(dense_tall, diag_square.T), 'dense t x diag s'
    assert compare(dense_tall, diag_tall.T), 'dense t x diag t'
    assert compare(dense_tall, diag_wide.T), 'dense t x diag w'
    assert compare(dense_wide, diag_square.T), 'dense w x diag s'
    assert compare(dense_wide, diag_tall.T), 'dense w x diag t'
    assert compare(dense_wide, diag_wide.T), 'dense w x diag w'

    # Test `B.matmul` with three matrices simultaneously.
    allclose(B.matmul(dense_tall, dense_square, dense_wide, tr_c=True),
             (to_np(dense_tall)
              .dot(to_np(dense_square))
              .dot(to_np(dense_wide).T)))

    # Test `Woodbury`.
    wb = lr + dense_tall
    assert compare(wb, dense_square.T), 'wb x dense s'
    assert compare(dense_square, wb.T), 'dense s x wb'
    assert compare(wb, wb.T), 'wb x wb'
    assert compare(wb.T, wb), 'wb x wb (2)'


def test_inverse_and_logdet():
    # Test `Dense`.
    a = np.random.randn(3, 3)
    a = Dense(a.dot(a.T))
    allclose(B.matmul(a, B.inverse(a)), np.eye(3))
    allclose(B.matmul(B.inverse(a), a), np.eye(3))
    allclose(B.logdet(a), np.log(np.linalg.det(to_np(a))))

    # Test `Diagonal`.
    d = Diagonal(np.array([1, 2, 3]))
    allclose(B.matmul(d, B.inverse(d)), np.eye(3))
    allclose(B.matmul(B.inverse(d), d), np.eye(3))
    allclose(B.logdet(d), np.log(np.linalg.det(to_np(d))))
    assert B.shape(B.inverse(Diagonal(np.array([1, 2]),
                                      rows=2, cols=4))) == (4, 2)

    # Test `Woodbury`.
    a = np.random.randn(3, 2)
    b = np.random.randn(2, 2) + 1e-2 * np.eye(2)
    wb = d + LowRank(left=a, middle=b.dot(b.T))
    for _ in range(4):
        allclose(B.matmul(wb, B.inverse(wb)), np.eye(3))
        allclose(B.matmul(B.inverse(wb), wb), np.eye(3))
        allclose(B.logdet(wb), np.log(np.linalg.det(to_np(wb))))
        wb = B.inverse(wb)

    # Test `LowRank`.
    with pytest.raises(RuntimeError):
        B.inverse(wb.lr)
    with pytest.raises(RuntimeError):
        B.logdet(wb.lr)


def test_lr_diff():
    # First, test correctness.
    a = np.random.randn(3, 2)
    b = np.random.randn(2, 2)
    lr1 = LowRank(left=a, right=np.random.randn(3, 2), middle=b.dot(b.T))
    a = np.random.randn(3, 2)
    b = np.random.randn(2, 2)
    lr2 = LowRank(left=a, right=np.random.randn(3, 2), middle=b.dot(b.T))

    allclose(B.lr_diff(lr1, lr1), B.zeros(3, 3))
    allclose(B.lr_diff(lr1 + lr2, lr1), lr2)
    allclose(B.lr_diff(lr1 + lr2, lr2), lr1)
    allclose(B.lr_diff(lr1 + lr1 + lr2, lr1), lr1 + lr2)
    allclose(B.lr_diff(lr1 + lr1 + lr2, lr2), lr1 + lr1)
    allclose(B.lr_diff(lr1 + lr1 + lr2, lr1 + lr1), lr2)
    allclose(B.lr_diff(lr1 + lr1 + lr2, lr1 + lr2), lr1)
    allclose(B.lr_diff(lr1 + lr1 + lr2, lr1 + lr1 + lr2), B.zeros(3, 3))

    # Second, test positive definiteness.
    lr1 = LowRank(left=lr1.left, middle=lr1.middle)
    lr2 = LowRank(left=lr2.left, middle=lr2.middle)

    B.cholesky(B.lr_diff(lr1, 0.999 * lr1))
    B.cholesky(B.lr_diff(lr1 + lr2, lr1))
    B.cholesky(B.lr_diff(lr1 + lr2, lr2))
    B.cholesky(B.lr_diff(lr1 + lr1 + lr2, lr1))
    B.cholesky(B.lr_diff(lr1 + lr1 + lr2, lr1 + lr1))
    B.cholesky(B.lr_diff(lr1 + lr1 + lr2, lr1 + lr2))
    B.cholesky(B.lr_diff(lr1 + lr1 + lr2, lr1 + lr1 + 0.999 * lr2))


def test_root():
    # Test `Dense`.
    a = np.random.randn(5, 5)
    a = Dense(a.dot(a.T))
    allclose(a, B.matmul(B.root(a), B.root(a)))

    # Test `Diagonal`.
    d = Diagonal(np.array([1, 2, 3, 4, 5]))
    allclose(d, B.matmul(B.root(d), B.root(d)))


def test_sample():
    a = np.random.randn(3, 3)
    a = Dense(a.dot(a.T))
    b = np.random.randn(2, 2)
    wb = Diagonal(B.diag(a)) + LowRank(left=np.random.randn(3, 2),
                                       middle=b.dot(b.T))

    # Test `Dense` and `Woodbury`.
    num_samps = 500000
    for cov in [a, wb]:
        samps = B.sample(cov, num_samps)
        cov_emp = B.matmul(samps, samps, tr_b=True) / num_samps
        assert np.mean(np.abs(to_np(cov_emp) - to_np(cov))) <= 5e-2


def test_schur():
    # Test `Dense`.
    a = np.random.randn(5, 10)
    b = np.random.randn(3, 5)
    c = np.random.randn(3, 3)
    d = np.random.randn(3, 10)
    c = c.dot(c.T)

    allclose(B.schur(a, b, c, d), a - np.linalg.solve(c.T, b).T.dot(d),
             desc='n n n n')

    # Test `Woodbury`.
    #   The inverse of the Woodbury matrix already properly tests the method for
    #   Woodbury matrices.
    c = np.random.randn(2, 2)
    c = Diagonal(np.array([1, 2, 3])) + LowRank(left=np.random.randn(3, 2),
                                                middle=c.dot(c.T))
    allclose(B.schur(a, b, c, d), a - np.linalg.solve(to_np(c).T, b).T.dot(d),
             desc='n n w n')

    #   Test all combinations of `Woodbury`, `LowRank`, and `Diagonal`.
    a = np.random.randn(2, 2)
    a = Diagonal(np.array([4, 5, 6, 7, 8]), rows=5, cols=10) + LowRank(
        left=np.random.randn(5, 2),
        right=np.random.randn(10, 2),
        middle=a.dot(a.T))

    b = np.random.randn(2, 2)
    b = Diagonal(np.array([9, 10, 11]), rows=3, cols=5) + LowRank(
        left=c.lr.left,
        right=a.lr.left,
        middle=b.dot(b.T))

    d = np.random.randn(2, 2)
    d = Diagonal(np.array([12, 13, 14]), rows=3, cols=10) + LowRank(
        left=c.lr.right,
        right=a.lr.right,
        middle=d.dot(d.T))

    #     Loop over all combinations. Some of them should be efficient and
    #     representation preserving; all of them should be correct.
    for ai in [a, a.lr, a.diag]:
        for bi in [b, b.lr, b.diag]:
            for ci in [c, c.diag]:
                for di in [d, d.lr, d.diag]:
                    allclose(B.schur(ai, bi, ci, di),
                             to_np(ai) -
                             np.linalg.solve(to_np(ci).T,
                                             to_np(bi)).T.dot(to_np(di)),
                             desc='{} {} {} {}'.format(ai, bi, ci, di))


def test_qf():
    # Generate some test inputs.
    b, c = np.random.randn(5, 3), np.random.randn(5, 3)

    # Generate some matrices to test.
    a = np.random.randn(5, 5)
    a = Dense(a.dot(a.T))
    d = Diagonal(B.diag(a))
    e = np.random.randn(2, 2)
    wb = d + LowRank(left=np.random.randn(5, 2),
                     middle=e.dot(e.T))

    for x in [a, d, wb]:
        allclose(B.qf(x, b), np.linalg.solve(to_np(x), b).T.dot(b))
        allclose(B.qf(x, b, b), B.qf(x, b))
        allclose(B.qf(x, b, c), np.linalg.solve(to_np(x), b).T.dot(c))
        allclose(B.qf_diag(x, b),
                 np.diag(np.linalg.solve(to_np(x), b).T.dot(b)))
        allclose(B.qf_diag(x, b, b), B.qf_diag(x, b, b))
        allclose(B.qf_diag(x, b, c),
                 np.diag(np.linalg.solve(to_np(x), b).T.dot(c)))

    # Test `LowRank`.
    lr = LowRank(np.random.randn(5, 3))
    with pytest.raises(RuntimeError):
        B.qf(lr, b)
    with pytest.raises(RuntimeError):
        B.qf(lr, b, c)
    with pytest.raises(RuntimeError):
        B.qf_diag(lr, b)
    with pytest.raises(RuntimeError):
        B.qf_diag(lr, b, c)


def test_ratio():
    a, b = np.random.randn(4, 4), np.random.randn(4, 4)
    a, b = Dense(a.dot(a.T)), Dense(b.dot(b.T))
    d, e = Diagonal(B.diag(a)), Diagonal(B.diag(b))
    c = np.random.randn(3, 3)
    lr = LowRank(left=np.random.randn(4, 3), middle=c.dot(c.T))

    allclose(B.ratio(a, b), np.trace(np.linalg.solve(to_np(b), to_np(a))))
    allclose(B.ratio(lr, b), np.trace(np.linalg.solve(to_np(b), to_np(lr))))
    allclose(B.ratio(d, e), np.trace(np.linalg.solve(to_np(e), to_np(d))))


def test_transposition():
    def compare(a):
        allclose(B.transpose(a), to_np(a).T)

    d = Diagonal(np.array([1, 2, 3]), rows=5, cols=10)
    lr = LowRank(left=np.random.randn(5, 3), right=np.random.randn(10, 3))

    compare(Dense(np.random.randn(5, 5)))
    compare(d)
    compare(lr)
    compare(d + lr)
    compare(Constant(5, rows=4, cols=2))


def test_arithmetic_and_shapes():
    a = Dense(np.random.randn(4, 3))
    d = Diagonal(np.array([1.0, 2.0, 3.0]), rows=4, cols=3)
    lr = LowRank(left=np.random.randn(4, 2),
                 right=np.random.randn(3, 2),
                 middle=np.random.randn(2, 2))
    zero = Zero.from_(a)
    one = One.from_(a)
    constant = Constant.from_(2.0, a)
    wb = d + lr

    # Aggregate all matrices.
    candidates = [a, d, lr, wb, constant, one, zero, 2, 1, 0, 2.0, 1.0, 0.0]

    # Check division.
    allclose(a.__div__(5.0), to_np(a) / 5.0)
    allclose(a.__rdiv__(5.0), 5.0 / to_np(a))
    allclose(a.__truediv__(5.0), to_np(a) / 5.0)
    allclose(a.__rtruediv__(5.0), 5.0 / to_np(a))
    allclose(B.divide(a, 5.0), to_np(a) / 5.0)
    allclose(B.divide(a, a), B.ones(to_np(a)))

    # Check shapes.
    for m in candidates:
        assert B.shape(a) == (4, 3)

    # Check interactions.
    for m1, m2 in product(candidates, candidates):
        allclose(m1 * m2, to_np(m1) * to_np(m2))
        allclose(m1 + m2, to_np(m1) + to_np(m2))
        allclose(m1 - m2, to_np(m1) - to_np(m2))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import numpy as np

from stheno.graph import Graph, GP
from stheno.input import MultiInput
from stheno.kernel import EQ
from stheno.momean import MultiOutputMean
from .util import allclose


def test_momean():
    m = Graph()
    p1 = GP(1 * EQ(), lambda x: 2 * x, graph=m)
    p2 = GP(2 * EQ().stretch(2), 1, graph=m)

    mom = MultiOutputMean(p1, p2)
    ms = m.means

    x = np.linspace(0, 1, 10)

    assert str(mom) == 'MultiOutputMean(<lambda>, 1)'

    allclose(mom(x), np.concatenate([ms[p1](x), ms[p2](x)], axis=0))
    allclose(mom(p1(x)), ms[p1](x))
    allclose(mom(p2(x)), ms[p2](x))
    allclose(mom(MultiInput(p2(x), p1(x))),
             np.concatenate([ms[p2](x), ms[p1](x)], axis=0))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import numpy as np
import pytest
import tensorflow as tf
from lab.tensorflow import B

from stheno.input import Observed, Unique
from stheno.kernel import (
    EQ,
    RQ,
    Matern12,
    Matern32,
    Matern52,
    Delta,
    FixedDelta,
    Kernel,
    Linear,
    OneKernel,
    ZeroKernel,
    PosteriorKernel,
    ShiftedKernel,
    TensorProductKernel,
    CorrectiveKernel,
    DecayingKernel,
    LogKernel,
    perturb,
)
from stheno.matrix import matrix, dense, Zero, One, UniformlyDiagonal
from .util import allclose


def standard_kernel_tests(k, shapes=None, dtype=np.float64):
    if shapes is None:
        shapes = [((10, 2), (5, 2)),
                  ((10, 1), (5, 1)),
                  ((10,), (5,)),
                  ((10,), ()),
                  ((), (5,)),
                  ((), ())]

    # Check various shapes of arguments.
    for shape1, shape2 in shapes:
        x1 = B.randn(dtype, *shape1)
        x2 = B.randn(dtype, *shape2)

        # Check that the kernel computes consistently.
        allclose(k(x1, x2), reversed(k)(x2, x1).T)

        # Check `elwise`.
        x2 = B.randn(dtype, *shape1)

        allclose(k.elwise(x1, x2)[:, 0], B.diag(k(x1, x2)))
        allclose(k.elwise(x1, x2), Kernel.elwise(k, x1, x2))
        # The element-wise computation is more accurate, which is why we allow
        # a discrepancy a bit larger than the square root of the machine
        # epsilon.
        allclose(k.elwise(x1)[:, 0], B.diag(k(x1)),
                 desc='', atol=1e-6, rtol=1e-6)
        allclose(k.elwise(x1), Kernel.elwise(k, x1))


def test_corner_cases():
    with pytest.raises(RuntimeError):
        Kernel()(1.)


def test_construction():
    k = EQ()

    x = np.random.randn(10, 1)

    k(x)
    k(x, x)

    k(Observed(x))
    k(Observed(x), Observed(x))
    k(x, Observed(x))
    k(Observed(x), x)

    k.elwise(x)
    k.elwise(x, x)

    k.elwise(Observed(x))
    k.elwise(Observed(x), Observed(x))
    k.elwise(x, Observed(x))
    k.elwise(Observed(x), x)


def test_basic_arithmetic():
    k1 = EQ()
    k2 = RQ(1e-1)
    k3 = Matern12()
    k4 = Matern32()
    k5 = Matern52()
    k6 = Delta()
    k7 = Linear()
    xs1 = np.random.randn(10, 2), np.random.randn(20, 2)
    xs2 = np.random.randn(), np.random.randn()

    allclose(k6(xs1[0]), k6(xs1[0], xs1[0]))
    allclose((k1 * k2)(*xs1), k1(*xs1) * k2(*xs1))
    allclose((k1 * k2)(*xs2), k1(*xs2) * k2(*xs2))
    allclose((k3 + k4)(*xs1), k3(*xs1) + k4(*xs1))
    allclose((k3 + k4)(*xs2), k3(*xs2) + k4(*xs2))
    allclose((5. * k5)(*xs1), 5. * k5(*xs1))
    allclose((5. * k5)(*xs2), 5. * k5(*xs2))
    allclose((5. + k7)(*xs1), 5. + k7(*xs1))
    allclose((5. + k7)(*xs2), 5. + k7(*xs2))
    allclose(k1.stretch(2.)(*xs1), k1(xs1[0] / 2., xs1[1] / 2.))
    allclose(k1.stretch(2.)(*xs2), k1(xs2[0] / 2., xs2[1] / 2.))
    allclose(k1.periodic(1.)(*xs1), k1.periodic(1.)(xs1[0], xs1[1] + 5.))
    allclose(k1.periodic(1.)(*xs2), k1.periodic(1.)(xs2[0], xs2[1] + 5.))


def test_reversal():
    x1 = np.random.randn(10, 2)
    x2 = np.random.randn(5, 2)
    x3 = np.random.randn()

    # Test with a stationary and non-stationary kernel.
    for k in [EQ(), Linear()]:
        allclose(k(x1), reversed(k)(x1))
        allclose(k(x3), reversed(k)(x3))
        allclose(k(x1, x2), reversed(k)(x1, x2))
        allclose(k(x1, x2), reversed(k)(x2, x1).T)

        # Test double reversal does the right thing.
        allclose(k(x1), reversed(reversed(k))(x1))
        allclose(k(x3), reversed(reversed(k))(x3))
        allclose(k(x1, x2), reversed(reversed(k))(x1, x2))
        allclose(k(x1, x2), reversed(reversed(k))(x2, x1).T)

    # Verify that the kernel has the right properties.
    k = reversed(EQ())
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 1
    assert k.period == np.inf

    k = reversed(Linear())
    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.var
    with pytest.raises(RuntimeError):
        k.length_scale
    assert k.period == np.inf
    assert str(k) == 'Reversed(Linear())'

    # Check equality.
    assert reversed(Linear()) == reversed(Linear())
    assert reversed(Linear()) != Linear()
    assert reversed(Linear()) != reversed(EQ())
    assert reversed(Linear()) != reversed(DecayingKernel(1, 1))

    # Standard tests:
    standard_kernel_tests(k)


def test_delta():
    k = Delta()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 0
    assert k.period == np.inf
    assert str(k) == 'Delta()'

    # Check equality.
    assert Delta() == Delta()
    assert Delta() != Delta(epsilon=k.epsilon * 10)
    assert Delta() != EQ()


@pytest.mark.parametrize('x1_x2', [(np.array(0), np.array(1)),
                                   (B.randn(10), B.randn(5)),
                                   (B.randn(10, 1), B.randn(5, 1)),
                                   (B.randn(10, 2), B.randn(5, 2))])
def test_delta_evaluations(x1_x2):
    k = Delta()
    x1, x2 = x1_x2
    n1 = B.shape(B.uprank(x1))[0]
    n2 = B.shape(B.uprank(x2))[0]

    # Check uniqueness checks.
    allclose(k(x1), B.eye(n1))
    allclose(k(x1, x2), B.zeros(n1, n2))

    # Standard tests:
    standard_kernel_tests(k)

    # Test `Unique` inputs.
    assert isinstance(k(Unique(x1), Unique(x1.copy())), Zero)
    assert isinstance(k(Unique(x1), Unique(x1)), UniformlyDiagonal)
    assert isinstance(k(Unique(x1), x1), Zero)
    assert isinstance(k(x1, Unique(x1)), Zero)

    assert isinstance(k.elwise(Unique(x1), Unique(x1.copy())), Zero)
    assert isinstance(k.elwise(Unique(x1), Unique(x1)), One)
    assert isinstance(k.elwise(Unique(x1), x1), Zero)
    assert isinstance(k.elwise(x1, Unique(x1)), Zero)


def test_fixed_delta():
    noises = B.rand(3)
    k = FixedDelta(noises)

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 0
    assert k.period == np.inf
    assert str(k) == 'FixedDelta()'

    # Check equality.
    assert FixedDelta(noises) == FixedDelta(noises)
    assert FixedDelta(noises) != FixedDelta(2 * noises)
    assert FixedDelta(noises) != EQ()

    # Standard tests:
    standard_kernel_tests(k)

    # Check correctness.
    x1 = B.randn(5)
    x2 = B.randn(5)
    allclose(k(x1), B.zeros(5, 5))
    allclose(k.elwise(x1), B.zeros(5, 1))
    allclose(k(x1, x2), B.zeros(5, 5))
    allclose(k.elwise(x1, x2), B.zeros(5, 1))

    x1 = B.randn(3)
    x2 = B.randn(3)
    allclose(k(x1), B.diag(noises))
    allclose(k.elwise(x1), B.uprank(noises))
    allclose(k(x1, x2), B.zeros(3, 3))
    allclose(k.elwise(x1, x2), B.zeros(3, 1))


def test_eq():
    k = EQ()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 1
    assert k.period == np.inf
    assert str(k) == 'EQ()'

    # Test equality.
    assert EQ() == EQ()
    assert EQ() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_rq():
    k = RQ(1e-1)

    # Verify that the kernel has the right properties.
    assert k.alpha == 1e-1
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 1
    assert k.period == np.inf
    assert str(k) == 'RQ(0.1)'

    # Test equality.
    assert RQ(1e-1) == RQ(1e-1)
    assert RQ(1e-1) != RQ(2e-1)
    assert RQ(1e-1) != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_exp():
    k = Matern12()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 1
    assert k.period == np.inf
    assert str(k) == 'Exp()'

    # Test equality.
    assert Matern12() == Matern12()
    assert Matern12() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_mat32():
    k = Matern32()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 1
    assert k.period == np.inf
    assert str(k) == 'Matern32()'

    # Test equality.
    assert Matern32() == Matern32()
    assert Matern32() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_mat52():
    k = Matern52()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 1
    assert k.period == np.inf
    assert str(k) == 'Matern52()'

    # Test equality.
    assert Matern52() == Matern52()
    assert Matern52() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_one():
    k = OneKernel()

    x1 = np.random.randn(10, 2)
    x2 = np.random.randn(5, 2)

    # Test that the kernel computes correctly.
    allclose(k(x1, x2), np.ones((10, 5)))

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == 0
    assert k.period == 0
    assert str(k) == '1'

    # Test equality.
    assert OneKernel() == OneKernel()
    assert OneKernel() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_zero():
    k = ZeroKernel()
    x1 = np.random.randn(10, 2)
    x2 = np.random.randn(5, 2)

    # Test that the kernel computes correctly.
    allclose(k(x1, x2), np.zeros((10, 5)))

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 0
    assert k.length_scale == 0
    assert k.period == 0
    assert str(k) == '0'

    # Test equality.
    assert ZeroKernel() == ZeroKernel()
    assert ZeroKernel() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_linear():
    k = Linear()

    # Verify that the kernel has the right properties.
    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.var
    with pytest.raises(RuntimeError):
        k.length_scale
    assert k.period == np.inf
    assert str(k) == 'Linear()'

    # Test equality.
    assert Linear() == Linear()
    assert Linear() != EQ()

    # Standard tests:
    standard_kernel_tests(k)


def test_decaying_kernel():
    k = DecayingKernel(3.0, 4.0)

    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.var
    assert k.period == np.inf
    assert str(k) == 'DecayingKernel(3.0, 4.0)'

    # Standard tests:
    standard_kernel_tests(k)

    # Test equality.
    assert DecayingKernel(3.0, 4.0) == DecayingKernel(3.0, 4.0)
    assert DecayingKernel(3.0, 4.0) != DecayingKernel(3.0, 5.0)
    assert DecayingKernel(3.0, 4.0) != DecayingKernel(4.0, 4.0)
    assert DecayingKernel(3.0, 4.0) != EQ()


def test_log_kernel():
    k = LogKernel()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert k.var == 1
    assert k.length_scale == np.inf
    assert k.period == np.inf
    assert str(k) == 'LogKernel()'

    # Test equality.
    assert LogKernel() == LogKernel()
    assert LogKernel() != EQ()

    # Standard tests:
    standard_kernel_tests(k)


def test_posterior_kernel():
    k = PosteriorKernel(
        EQ(), EQ(), EQ(),
        np.random.randn(5, 2), matrix(EQ()(np.random.randn(5, 1)))
    )

    # Verify that the kernel has the right properties.
    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.var
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.period
    assert str(k) == 'PosteriorKernel()'

    # Standard tests:
    standard_kernel_tests(k, shapes=[((10, 2), (5, 2))])


def test_corrective_kernel():
    a, b = np.random.randn(3, 3), np.random.randn(3, 3)
    a, b = a.dot(a.T), b.dot(b.T)
    z = np.random.randn(3, 2)
    k = CorrectiveKernel(EQ(), EQ(), z, a, matrix(b))

    # Verify that the kernel has the right properties.
    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.var
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.period
    assert str(k) == 'CorrectiveKernel()'

    # Standard tests:
    standard_kernel_tests(k, shapes=[((10, 2), (5, 2))])


def test_sum():
    k1 = EQ().stretch(2)
    k2 = 3 * RQ(1e-2).stretch(5)
    k = k1 + k2

    assert k.stationary
    allclose(k.length_scale, (1 * 2 + 3 * 5) / 4)
    assert k.period == np.inf
    assert k.var == 4

    allclose((EQ() + EQ()).length_scale, 1)
    allclose((EQ().stretch(2) + EQ().stretch(2)).length_scale, 2)

    assert EQ() + Linear() == EQ() + Linear()
    assert EQ() + Linear() == Linear() + EQ()
    assert EQ() + Linear() != EQ() + RQ(1e-1)
    assert EQ() + Linear() != RQ(1e-1) + Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_stretched():
    k = EQ().stretch(2)

    assert k.stationary
    assert k.length_scale == 2
    assert k.period == np.inf
    assert k.var == 1

    # Test equality.
    assert EQ().stretch(2) == EQ().stretch(2)
    assert EQ().stretch(2) != EQ().stretch(3)
    assert EQ().stretch(2) != Matern12().stretch(2)

    # Standard tests:
    standard_kernel_tests(k)

    k = EQ().stretch(1, 2)

    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.period
    assert k.var == 1

    # Check passing in a list.
    k = EQ().stretch(np.array([1, 2]))
    k(np.random.randn(10, 2))


def test_periodic():
    k = EQ().stretch(2).periodic(3)

    assert k.stationary
    assert k.length_scale == 2
    assert k.period == 3
    assert k.var == 1

    # Test equality.
    assert EQ().periodic(2) == EQ().periodic(2)
    assert EQ().periodic(2) != EQ().periodic(3)
    assert Matern12().periodic(2) != EQ().periodic(2)

    # Standard tests:
    standard_kernel_tests(k)

    k = 5 * k.stretch(5)

    assert k.stationary
    assert k.length_scale == 10
    assert k.period == 15
    assert k.var == 5

    # Check passing in a list.
    k = EQ().periodic(np.array([1, 2]))
    k(np.random.randn(10, 2))


def test_scaled():
    k = 2 * EQ()

    assert k.stationary
    assert k.length_scale == 1
    assert k.period == np.inf
    assert k.var == 2

    # Test equality.
    assert 2 * EQ() == 2 * EQ()
    assert 2 * EQ() != 3 * EQ()
    assert 2 * EQ() != 2 * Matern12()

    # Standard tests:
    standard_kernel_tests(k)


def test_shifted():
    k = ShiftedKernel(2 * EQ(), 5)

    assert k.stationary
    assert k.length_scale == 1
    assert k.period == np.inf
    assert k.var == 2

    # Test equality.
    assert Linear().shift(2) == Linear().shift(2)
    assert Linear().shift(2) != Linear().shift(3)
    assert Linear().shift(2) != DecayingKernel(1, 1).shift(2)

    # Standard tests:
    standard_kernel_tests(k)

    k = (2 * EQ()).shift(5, 6)

    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.length_scale
    assert k.period == np.inf
    assert k.var == 2

    # Check computation.
    x1 = np.random.randn(10, 2)
    x2 = np.random.randn(5, 2)
    k = Linear()
    allclose(k.shift(5)(x1, x2), k(x1 - 5, x2 - 5))

    # Check passing in a list.
    k = Linear().shift(np.array([1, 2]))
    k(np.random.randn(10, 2))


def test_product():
    k = (2 * EQ().stretch(10)) * (3 * RQ(1e-2).stretch(20))

    assert k.stationary
    assert k.length_scale == 10
    assert k.period == np.inf
    assert k.var == 6

    # Test equality.
    assert EQ() * Linear() == EQ() * Linear()
    assert EQ() * Linear() == Linear() * EQ()
    assert EQ() * Linear() != EQ() * RQ(1e-1)
    assert EQ() * Linear() != RQ(1e-1) * Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_selection():
    k = (2 * EQ().stretch(5)).select(0)

    assert k.stationary
    assert k.length_scale == 5
    assert k.period == np.inf
    assert k.var == 2

    # Test equality.
    assert EQ().select(0) == EQ().select(0)
    assert EQ().select(0) != EQ().select(1)
    assert EQ().select(0) != Matern12().select(0)

    # Standard tests:
    standard_kernel_tests(k)

    k = (2 * EQ().stretch(5)).select([2, 3])

    assert k.stationary
    assert k.length_scale == 5
    assert k.period == np.inf
    assert k.var == 2

    k = (2 * EQ().stretch(np.array([1, 2, 3]))).select([0, 2])

    assert k.stationary
    allclose(k.length_scale, [1, 3])
    allclose(k.period, [np.inf, np.inf])
    assert k.var == 2

    k = (2 * EQ().periodic(np.array([1, 2, 3]))).select([1, 2])

    assert k.stationary
    allclose(k.length_scale, [1, 1])
    allclose(k.period, [2, 3])
    assert k.var == 2

    k = (2 * EQ().stretch(np.array([1, 2, 3]))).select([0, 2], [1, 2])

    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.period
    assert k.var == 2

    k = (2 * EQ().periodic(np.array([1, 2, 3]))).select([0, 2], [1, 2])

    assert not k.stationary
    assert k.length_scale == 1
    with pytest.raises(RuntimeError):
        k.period
    assert k.var == 2

    # Test that computation is valid.
    k1 = EQ().select([1, 2])
    k2 = EQ()
    x = np.random.randn(10, 3)
    allclose(k1(x), k2(x[:, [1, 2]]))


def test_input_transform():
    k = Linear().transform(lambda x: x - 5)

    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.var
    with pytest.raises(RuntimeError):
        k.period

    def f1(x):
        return x

    def f2(x):
        return x ** 2

    # Test equality.
    assert EQ().transform(f1) == EQ().transform(f1)
    assert EQ().transform(f1) != EQ().transform(f2)
    assert EQ().transform(f1) != Matern12().transform(f1)

    # Standard tests:
    standard_kernel_tests(k)

    # Test computation of the kernel.
    k = Linear()
    x1, x2 = np.random.randn(10, 2), np.random.randn(10, 2)

    k2 = k.transform(lambda x: x ** 2)
    k3 = k.transform(lambda x: x ** 2, lambda x: x - 5)

    allclose(k(x1 ** 2, x2 ** 2), k2(x1, x2))
    allclose(k(x1 ** 2, x2 - 5), k3(x1, x2))


def test_tensor_product():
    k = TensorProductKernel(lambda x: B.sum(x ** 2, axis=1))

    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.var
    with pytest.raises(RuntimeError):
        k.period

    # Test equality.
    assert k == k
    assert k != TensorProductKernel(lambda x: x)
    assert k != EQ()

    # Standard tests:
    standard_kernel_tests(k)

    # Check computation of kernel.
    k = TensorProductKernel(lambda x: x)
    x1 = np.linspace(0, 1, 100)[:, None]
    x2 = np.linspace(0, 1, 50)[:, None]

    allclose(k(x1), x1 * x1.T)
    allclose(k(x1, x2), x1 * x2.T)

    k = TensorProductKernel(lambda x: x ** 2)

    allclose(k(x1), x1 ** 2 * (x1 ** 2).T)
    allclose(k(x1, x2), (x1 ** 2) * (x2 ** 2).T)


def test_derivative():
    # First, check properties.
    k = EQ().diff(0)

    assert not k.stationary
    with pytest.raises(RuntimeError):
        k.length_scale
    with pytest.raises(RuntimeError):
        k.var
    with pytest.raises(RuntimeError):
        k.period

    # Test equality.
    assert EQ().diff(0) == EQ().diff(0)
    assert EQ().diff(0) != EQ().diff(1)
    assert Matern12().diff(0) != EQ().diff(0)

    # Standard tests:
    for k in [EQ().diff(0), EQ().diff(None, 0), EQ().diff(0, None)]:
        standard_kernel_tests(k, dtype=tf.float64)

    # Check that a derivative must be specified.
    with pytest.raises(RuntimeError):
        EQ().diff(None, None)(np.array([1.0]))
    with pytest.raises(RuntimeError):
        EQ().diff(None, None).elwise(np.array([1.0]))


def test_derivative_eq():
    # Test derivative of kernel `EQ()`.
    k = EQ()
    x1 = B.randn(tf.float64, 10, 1)
    x2 = B.randn(tf.float64, 5, 1)

    # Test derivative with respect to first input.
    allclose(k.diff(0, None)(x1, x2), -k(x1, x2) * (x1 - B.transpose(x2)))
    allclose(k.diff(0, None)(x1), -k(x1) * (x1 - B.transpose(x1)))

    # Test derivative with respect to second input.
    allclose(k.diff(None, 0)(x1, x2), -k(x1, x2) * (B.transpose(x2) - x1))
    allclose(k.diff(None, 0)(x1), -k(x1) * (B.transpose(x1) - x1))

    # Test derivative with respect to both inputs.
    ref = k(x1, x2) * (1 - (x1 - B.transpose(x2)) ** 2)
    allclose(k.diff(0, 0)(x1, x2), ref)
    allclose(k.diff(0)(x1, x2), ref)
    ref = k(x1) * (1 - (x1 - B.transpose(x1)) ** 2)
    allclose(k.diff(0, 0)(x1), ref)
    allclose(k.diff(0)(x1), ref)


def test_derivative_linear():
    # Test derivative of kernel `Linear()`.
    k = Linear()
    x1 = B.randn(tf.float64, 10, 1)
    x2 = B.randn(tf.float64, 5, 1)

    # Test derivative with respect to first input.
    allclose(k.diff(0, None)(x1, x2),
             B.ones(tf.float64, 10, 5) * B.transpose(x2))
    allclose(k.diff(0, None)(x1),
             B.ones(tf.float64, 10, 10) * B.transpose(x1))

    # Test derivative with respect to second input.
    allclose(k.diff(None, 0)(x1, x2), B.ones(tf.float64, 10, 5) * x1)
    allclose(k.diff(None, 0)(x1), B.ones(tf.float64, 10, 10) * x1)

    # Test derivative with respect to both inputs.
    ref = B.ones(tf.float64, 10, 5)
    allclose(k.diff(0, 0)(x1, x2), ref)
    allclose(k.diff(0)(x1, x2), ref)
    ref = B.ones(tf.float64, 10, 10)
    allclose(k.diff(0, 0)(x1), ref)
    allclose(k.diff(0)(x1), ref)


@pytest.mark.parametrize('x,result',
                         [(np.float64([1]), np.float64([1e-20 + 1 + 1e-14])),
                          (np.float32([1]), np.float32([1e-20 + 1 + 1e-7]))])
def test_perturb(x, result):
    assert perturb(x) == result  # Test NumPy.
    assert perturb(tf.constant(x)).numpy() == result  # Test TF.


def test_perturb_type_check():
    with pytest.raises(ValueError):
        perturb(0)


@pytest.mark.parametrize('dtype', [tf.float32, tf.float64])
def test_nested_derivatives(dtype):
    x = B.randn(dtype, 10, 2)

    res = EQ().diff(0, 0).diff(0, 0)(x)
    assert ~B.isnan(res[0, 0])

    res = EQ().diff(1, 1).diff(1, 1)(x)
    assert ~B.isnan(res[0, 0])



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import numpy as np
import pytest

from stheno.graph import Graph, GP
from stheno.input import MultiInput
from stheno.kernel import EQ
from stheno.matrix import dense
from stheno.mokernel import MultiOutputKernel
from .util import allclose


def test_mokernel():
    m = Graph()
    p1 = GP(1 * EQ(), graph=m)
    p2 = GP(2 * EQ().stretch(2), graph=m)

    mok = MultiOutputKernel(p1, p2)
    ks = m.kernels

    x1 = np.linspace(0, 1, 10)
    x2 = np.linspace(1, 2, 5)
    x3 = np.linspace(1, 2, 10)

    assert str(mok) == 'MultiOutputKernel(EQ(), 2 * (EQ() > 2))'

    # `B.Numeric` versus `B.Numeric`:
    allclose(mok(x1, x2),
             np.concatenate([np.concatenate([dense(ks[p1, p1](x1, x2)),
                                             dense(ks[p1, p2](x1, x2))],
                                            axis=1),
                             np.concatenate([dense(ks[p2, p1](x1, x2)),
                                             dense(ks[p2, p2](x1, x2))],
                                            axis=1)], axis=0))
    allclose(mok.elwise(x1, x3),
             np.concatenate([ks[p1, p1].elwise(x1, x3),
                             ks[p2, p2].elwise(x1, x3)], axis=0))

    # `B.Numeric` versus `At`:
    allclose(mok(p1(x1), x2),
             np.concatenate([dense(ks[p1, p1](x1, x2)),
                             dense(ks[p1, p2](x1, x2))], axis=1))
    allclose(mok(p2(x1), x2),
             np.concatenate([dense(ks[p2, p1](x1, x2)),
                             dense(ks[p2, p2](x1, x2))], axis=1))
    allclose(mok(x1, p1(x2)),
             np.concatenate([dense(ks[p1, p1](x1, x2)),
                             dense(ks[p2, p1](x1, x2))], axis=0))
    allclose(mok(x1, p2(x2)),
             np.concatenate([dense(ks[p1, p2](x1, x2)),
                             dense(ks[p2, p2](x1, x2))], axis=0))
    with pytest.raises(ValueError):
        mok.elwise(x1, p2(x3))
    with pytest.raises(ValueError):
        mok.elwise(p1(x1), x3)

    # `At` versus `At`:
    allclose(mok(p1(x1), p1(x2)), ks[p1](x1, x2))
    allclose(mok(p1(x1), p2(x2)), ks[p1, p2](x1, x2))
    allclose(mok.elwise(p1(x1), p1(x3)), ks[p1].elwise(x1, x3))
    allclose(mok.elwise(p1(x1), p2(x3)), ks[p1, p2].elwise(x1, x3))

    # `MultiInput` versus `MultiInput`:
    allclose(mok(MultiInput(p2(x1), p1(x2)), MultiInput(p2(x1))),
             np.concatenate([dense(ks[p2, p2](x1, x1)),
                             dense(ks[p1, p2](x2, x1))], axis=0))
    with pytest.raises(ValueError):
        mok.elwise(MultiInput(p2(x1), p1(x3)), MultiInput(p2(x1)))
    allclose(mok.elwise(MultiInput(p2(x1), p1(x3)),
                        MultiInput(p2(x1), p1(x3))),
             np.concatenate([ks[p2, p2].elwise(x1, x1),
                             ks[p1, p1].elwise(x3, x3)], axis=0))

    # `MultiInput` versus `At`:
    allclose(mok(MultiInput(p2(x1), p1(x2)), p2(x1)),
             np.concatenate([dense(ks[p2, p2](x1, x1)),
                             dense(ks[p1, p2](x2, x1))], axis=0))
    allclose(mok(p2(x1), MultiInput(p2(x1), p1(x2))),
             np.concatenate([dense(ks[p2, p2](x1, x1)),
                             dense(ks[p2, p1](x1, x2))], axis=1))
    with pytest.raises(ValueError):
        mok.elwise(MultiInput(p2(x1), p1(x3)), p2(x1))
    with pytest.raises(ValueError):
        mok.elwise(p2(x1), MultiInput(p2(x1), p1(x3)))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from time import time

import numpy as np
from lab import B
from plum import dispatch
from numpy.testing import assert_array_almost_equal

from stheno.matrix import Dense, dense as _dense

__all__ = ['benchmark', 'to_np', 'allclose', 'approx']


def benchmark(f, args, n=1000, get_output=False):
    """Benchmark the performance of a function `f` called with arguments
    `args` in microseconds.

    Args:
        f (function): Function to benchmark.
        args (tuple): Argument to call `f` with.
        n (int): Repetitions.
        get_output (bool): Also return final output of function.
    """
    start = time()
    for i in range(n):
        out = f(*args)
    dur = (time() - start) * 1e6 / n
    return (dur, out) if get_output else out


@dispatch(B.Numeric)
def to_np(a):
    return a


@dispatch(tuple)
def to_np(a):
    return tuple(to_np(x) for x in a)


@dispatch(Dense)
def to_np(a):
    return _dense(a)


@dispatch(list)
def to_np(a):
    return np.array(a)


@dispatch({B.TFNumeric, B.TorchNumeric})
def to_np(a):
    return a.numpy()


@dispatch({B.Numeric, Dense, list}, {B.Numeric, Dense, list}, [object])
def allclose(a, b, desc=None, atol=1e-8, rtol=1e-8):
    np.testing.assert_allclose(to_np(a), to_np(b),
                               atol=atol, rtol=rtol, err_msg=desc)


@dispatch(tuple, tuple, [object])
def allclose(a, b, desc=None, atol=1e-8, rtol=1e-8):
    assert len(a) == len(b)
    for x, y in zip(a, b):
        allclose(x, y, desc, atol=atol, rtol=rtol)


approx = assert_array_almost_equal



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import logging

from lab import B
from plum import Dispatcher, Self, Referentiable, type_parameter

from .input import At, MultiInput
from .mean import Mean

__all__ = ['MultiOutputMean']

log = logging.getLogger(__name__)


class MultiOutputMean(Mean, Referentiable):
    """A generic multi-output mean.

    Args:
        *ps (instance of :class:`.graph.GP`): Processes that make up the
            multi-valued process.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, *ps):
        self.means = ps[0].graph.means
        self.ps = ps

    @_dispatch(B.Numeric)
    def __call__(self, x):
        return self(MultiInput(*(p(x) for p in self.ps)))

    @_dispatch(At)
    def __call__(self, x):
        return self.means[type_parameter(x)](x.get())

    @_dispatch(MultiInput)
    def __call__(self, x):
        return B.concat(*[self(xi) for xi in x.get()], axis=0)

    def __str__(self):
        ks = [str(self.means[p]) for p in self.ps]
        return 'MultiOutputMean({})'.format(', '.join(ks))



# -*- coding: utf-8 -*-

from __future__ import absolute_import  # pragma: no cover

import logging

# noinspection PyUnresolvedReferences
import lab.tensorflow
from plum import Dispatcher
from tensorflow import Tensor

# noinspection PyUnresolvedReferences
from . import *  # pragma: no cover
from .field import Element
from .random import Random

log = logging.getLogger(__name__)

_dispatch = Dispatcher()

# Save original methods.
__mul__tensor = Tensor.__mul__
__add__tensor = Tensor.__add__


@_dispatch(Tensor, object)
def __mul__(self, other):
    return __mul__tensor(self, other)


@_dispatch(Tensor, {Element, Random})
def __mul__(self, other):
    return other.__rmul__(self)


@_dispatch(Tensor, object)
def __add__(self, other):
    return __add__tensor(self, other)


@_dispatch(Tensor, {Element, Random})
def __add__(self, other):
    return other.__radd__(self)


# Assign dispatchable methods.
Tensor.__mul__ = __mul__
Tensor.__add__ = __add__
